# **************************************************************************
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import  logging
logger = logging.getLogger(__name__)

import pyworkflow.tests as pwtests

from pwem import Domain
import pwem.protocols as emprot

from .test_workflow import TestWorkflow


class TestMixedBPV(TestWorkflow):
    @classmethod
    def setUpClass(cls):
        pwtests.setupTestProject(cls)
        cls.dataset = pwtests.DataSet.getDataSet('xmipp_tutorial')
        cls.crdsDir = cls.dataset.getFile('boxingDir')
        cls.micsFn = cls.dataset.getFile('allMics')
        cls.vol1 = cls.dataset.getFile('vol1')

    def test_workflow(self):
        try:
            from itertools import izip
        except ImportError:
            izip = zip
        # First, import a set of micrographs
        protImport = self.newProtocol(emprot.ProtImportMicrographs,
                                      filesPath=self.micsFn,
                                      samplingRate=1.237, voltage=300)
        self.launchProtocol(protImport)
        self.assertIsNotNone(protImport.outputMicrographs,
                             "There was a problem with the import")
        #         self.validateFiles('protImport', protImport)

        # Import a set of volumes
        logger.info("Import Volume")
        protImportVol = self.newProtocol(emprot.ProtImportVolumes, filesPath=self.vol1,
                                         samplingRate=9.896)
        self.launchProtocol(protImportVol)
        self.assertIsNotNone(protImportVol.getFiles(),
                             "There was a problem with the import")
        #        self.validateFiles('protImportVol', protImportVol)

        # Perform a downsampling on the micrographs
        logger.info("Downsampling...")
        XmippProtPreprocessMicrographs = Domain.importFromPlugin('xmipp3.protocols',
                                                                 'XmippProtPreprocessMicrographs',
                                                                 doRaise=True)
        protDownsampling = self.newProtocol(XmippProtPreprocessMicrographs,
                                            doDownsample=True, downFactor=5,
                                            doCrop=False, runMode=1)
        protDownsampling.inputMicrographs.set(protImport.outputMicrographs)
        self.launchProtocol(protDownsampling)
        self.assertIsNotNone(protDownsampling.outputMicrographs,
                             "There was a problem with the downsampling")
        #         self.validateFiles('protDownsampling', protDownsampling)

        # Estimate CTF on the downsampled micrographs
        logger.info("Performing CTFfind...")
        ProtCTFFind = Domain.importFromPlugin('cistem.protocols',
                                              'CistemProtCTFFind', doRaise=True)
        protCTF = self.newProtocol(ProtCTFFind, numberOfThreads=4,
                                   minDefocus=22000, maxDefocus=25000)
        protCTF.inputMicrographs.set(protDownsampling.outputMicrographs)
        self.launchProtocol(protCTF)
        self.assertIsNotNone(protCTF.outputCTF,
                             "There was a problem with the CTF estimation")

        valuesList = [[24000, 24000], [22548, 22518], [23058, 23029]]
        for ctfModel, values in izip(protCTF.outputCTF, valuesList):
            self.assertAlmostEquals(ctfModel.getDefocusU(), values[0], delta=1000)
            self.assertAlmostEquals(ctfModel.getDefocusV(), values[1], delta=1000)
            self.assertAlmostEquals(ctfModel.getMicrograph().getSamplingRate(),
                                    6.185, delta=0.001)

        #         self.validateFiles('protCTF', protCTF)

        logger.info("Running Eman import coordinates...")
        protPP = self.newProtocol(emprot.ProtImportCoordinates,
                                  importFrom=emprot.ProtImportCoordinates.IMPORT_FROM_EMAN,
                                  filesPath=self.crdsDir,
                                  filesPattern='*_info.json', boxSize=110)
        protPP.inputMicrographs.set(protDownsampling.outputMicrographs)
        self.launchProtocol(protPP)
        self.assertIsNotNone(protPP.outputCoordinates,
                             "There was a problem with the Eman import coordinates")

        # Extract the SetOfParticles.
        logger.info("Run extract particles with other downsampling factor")
        XmippProtExtractParticles = Domain.importFromPlugin(
            'xmipp3.protocols',
            'XmippProtExtractParticles')
        OTHER = Domain.importFromPlugin('xmipp3.constants', 'OTHER')
        protExtract = self.newProtocol(XmippProtExtractParticles,
                                       boxSize=64,
                                       downsampleType=OTHER,
                                       downFactor=8.0,
                                       doFlip=False, runMode=1, doInvert=False)
        protExtract.inputCoordinates.set(protPP.outputCoordinates)
        protExtract.ctfRelations.set(protCTF.outputCTF)
        protExtract.inputMicrographs.set(protImport.outputMicrographs)
        self.launchProtocol(protExtract)
        self.assertIsNotNone(protExtract.outputParticles,
                             "There was a problem with the extract particles")
        #         self.validateFiles('protExtract', protExtract)


class TestMixedBPV2(TestWorkflow):

    @classmethod
    def setUpClass(cls):
        pwtests.setupTestProject(cls)
        cls.dataset = pwtests.DataSet.getDataSet('xmipp_tutorial')
        cls.crdsDir = cls.dataset.getFile('boxingDir')
        cls.micsFn = cls.dataset.getFile('allMics')
        cls.vol1 = cls.dataset.getFile('vol1')

    def test_workflow(self):
        # First, import a set of micrographs
        protImport = self.newProtocol(emprot.ProtImportMicrographs,
                                      filesPath=self.micsFn,
                                      samplingRate=1.237, voltage=300)
        self.launchProtocol(protImport)
        self.assertIsNotNone(protImport.outputMicrographs,
                             "There was a problem with the import")
        #         self.validateFiles('protImport', protImport)

        # Import a set of volumes
        logger.info("Import Volume")
        protImportVol = self.newProtocol(emprot.ProtImportVolumes, filesPath=self.vol1,
                                         samplingRate=9.896)
        self.launchProtocol(protImportVol)
        self.assertIsNotNone(protImportVol.getFiles(),
                             "There was a problem with the import")
        #        self.validateFiles('protImportVol', protImportVol)

        # Perform a downsampling on the micrographs
        logger.info("Downsampling...")
        XmippProtPreprocessMicrographs = Domain.importFromPlugin(
            'xmipp3.protocols',
            'XmippProtPreprocessMicrographs',
            doRaise=True)
        protDownsampling = self.newProtocol(XmippProtPreprocessMicrographs,
                                            doDownsample=True, downFactor=5,
                                            doCrop=False, runMode=1)
        protDownsampling.inputMicrographs.set(protImport.outputMicrographs)
        self.launchProtocol(protDownsampling)
        self.assertIsNotNone(protDownsampling.outputMicrographs,
                             "There was a problem with the downsampling")
        #         self.validateFiles('protDownsampling', protDownsampling)

        # Estimate CTF on the downsampled micrographs
        logger.info("Performing CTFfind...")
        ProtCTFFind = Domain.importFromPlugin('cistem.protocols',
                                              'CistemProtCTFFind', doRaise=True)
        protCTF = self.newProtocol(ProtCTFFind, numberOfThreads=4,
                                   minDefocus=22000, maxDefocus=25000)
        protCTF.inputMicrographs.set(protImport.outputMicrographs)
        self.launchProtocol(protCTF)
        self.assertIsNotNone(protCTF.outputCTF,
                             "There was a problem with the CTF estimation")
        #         self.validateFiles('protCTF', protCTF)

        logger.info("Running Eman import coordinates...")
        protPP = self.newProtocol(emprot.ProtImportCoordinates,
                                  importFrom=emprot.ProtImportCoordinates.IMPORT_FROM_EMAN,
                                  filesPath=self.crdsDir,
                                  filesPattern='*_info.json', boxSize=110)
        protPP.inputMicrographs.set(protDownsampling.outputMicrographs)
        self.launchProtocol(protPP)
        self.assertIsNotNone(protPP.outputCoordinates,
                             "There was a problem with the Eman import coordinates")

        logger.info("<Run extract particles with Same as picking>")
        XmippProtExtractParticles = Domain.importFromPlugin(
            'xmipp3.protocols',
            'XmippProtExtractParticles')

        SAME_AS_PICKING = Domain.importFromPlugin('xmipp3.constants',
                                                  'SAME_AS_PICKING')
        protExtract = self.newProtocol(XmippProtExtractParticles, boxSize=110,
                                       downsampleType=SAME_AS_PICKING,
                                       doFlip=True, doInvert=True, runMode=1)
        protExtract.inputCoordinates.set(protPP.outputCoordinates)
        protExtract.ctfRelations.set(protCTF.outputCTF)
        self.launchProtocol(protExtract)
        self.assertIsNotNone(protExtract.outputParticles,
                             "There was a problem with the extract particles")
        # self.validateFiles('protExtract', protExtract)

        logger.info("Run Preprocess particles")
        XmippProtCropResizeParticles = Domain.importFromPlugin(
            'xmipp3.protocols',
            'XmippProtCropResizeParticles')
        protCropResize = self.newProtocol(XmippProtCropResizeParticles,
                                          doResize=True, resizeOption=1,
                                          resizeDim=110)
        protCropResize.inputParticles.set(protExtract.outputParticles)
        self.launchProtocol(protCropResize)

        self.assertIsNotNone(protCropResize.outputParticles,
                             "There was a problem with resize/crop the particles")

        logger.info("Run ML2D")
        XmippProtML2D = Domain.importFromPlugin('xmipp3.protocols',
                                                'XmippProtML2D')
        protML2D = self.newProtocol(XmippProtML2D, numberOfClasses=8, maxIters=2,
                                    numberOfMpi=2, numberOfThreads=2)
        protML2D.inputParticles.set(protCropResize.outputParticles)
        self.launchProtocol(protML2D)
        self.assertIsNotNone(protML2D.outputClasses,
                             "There was a problem with ML2D")
        # self.validateFiles('protML2D', protML2D)

        #         #FIXME: Check the initial model with EMAn and restore the next step
        #         return

        logger.info("Run Initial Model")

        EmanProtInitModel = Domain.importFromPlugin('eman2.protocols',
                                                    'EmanProtInitModel',
                                                    doRaise=True)
        protIniModel = self.newProtocol(EmanProtInitModel, numberOfIterations=1,
                                        numberOfModels=2,
                                        shrink=5, symmetry='icos',
                                        numberOfThreads=3)
        protIniModel.inputSet.set(protML2D.outputClasses)
        self.launchProtocol(protIniModel)
        self.assertIsNotNone(protIniModel.outputVolumes,
                             "There was a problem with Initial Model")



# ***************************************************************************
# * Authors:    Roberto Marabini (roberto@cnb.csic.es)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# ***************************************************************************/
import os
import time
from urllib.request import urlopen
import json

import pyworkflow.webservices as pws
import pyworkflow.tests as pwtests

import pwem.protocols as emprot

# Class intentionally renamed to avoid muting this test

class NotifierTest(pwtests.BaseTest):
    @classmethod
    def setUpClass(cls):
        pwtests.setupTestProject(cls)

    def _getUrl(self):
        return os.environ.get('SCIPION_NOTIFY_URL',
                              pws.SCIPION_STATS_WORKFLOW_APP).strip()

    def test_projectNotifier(self):
        """ Execute a protocol and then report on it
        """
        # Connect to heroku and retrieve protocol list
        # then store number of times protstress has been executed
        url = self._getUrl()

        # Change periodicity in notification
        os.environ["SCIPION_NOTIFY_SECONDS"] = "30"
        os.environ["SCIPION_NOTIFY"] = "True"
        url = url.replace("workflow/workflow/",
                          "workflow/protocol/?name=ProtStress")
        results = json.loads(urlopen(url).read().decode('utf-8'))
        objects = results["objects"]
        if len(objects) != 0:
            objects = results["objects"][0]
            times_protocolRemote = objects["timesUsed"]
        else:
            times_protocolRemote = 0

        # Run a project that executes one time protStress
        kwargs = {'noCpu': 2,
                  'noMem': 0,
                  'amountMem': 8,
                  'timeout': 3,
                  'delay': 1
                  }
        # Create and execute protocol stress
        prot1 = self.newProtocol(emprot.ProtStress, **kwargs)
        prot1.setObjLabel('stress')
        self.proj.launchProtocol(prot1, wait=True)

        # We want to test this class: ProjectNotifier
        projectNotifier = pws.ProjectWorkflowNotifier(self.proj)
        # Remove uuid and data files (so we start always in the same conditions)
        uuidFn = projectNotifier._getUuidFileName()
        os.remove(uuidFn) if os.path.exists(uuidFn) else None
        dataFn = projectNotifier._getDataFileName()
        os.remove(dataFn) if os.path.exists(dataFn) else None

        # Store workflow  in database
        projectNotifier.notifyWorkflow()
        # Wait until protocol gui has connected to heroku and sent the information
        time.sleep(5)
        # Get uuid from local file
        uuid = projectNotifier._getUuid()
        # Get protocol list from database
        urlWork = self._getUrl()
        urlWork += "?project_uuid=" + uuid
        results = json.loads(urlopen(urlWork).read().decode('utf-8'))

        objects = results["objects"]
        if len(objects) != 0:
            objects = results["objects"][0]
            project_workflowRemote = objects["project_workflow"]
        else:
            project_workflowRemote = ""
        # Get protocol list local
        project_workflowLocal = self.proj.getProtocolsJson(namesOnly=True)
        # Test that stored protocol and local one are identical
        self.assertEqual(project_workflowLocal, project_workflowRemote)

        # Now check that the  number of times postgress has been executed
        # has increased in 1
        urlProt = self._getUrl()
        urlProt = urlProt.replace("workflow/workflow/",
                                  "workflow/protocol/?name=ProtStress")
        time.sleep(5)  # notifier runs in a thread so wait a bit
        results = json.loads(urlopen(urlProt).read().decode('utf-8'))
        objects = results["objects"]
        if len(objects) != 0:
            objects = results["objects"][0]
            times_protocolRemote_2 = objects["timesUsed"]
        else:
            times_protocolRemote_2 = 0
        self.assertEqual(times_protocolRemote_2, times_protocolRemote + 1)

        # Try to resend the project, as we have a 30 sec time this should
        # not go through number times should not change
        projectNotifier.notifyWorkflow()
        time.sleep(5)  # notifier runs in a thread so wait a bit
        results = json.loads(urlopen(urlProt).read().decode('utf-8'))
        objects = results["objects"]
        if len(objects) != 0:
            objects = results["objects"][0]
            times_protocolRemote_2 = objects["timesUsed"]
        else:
            times_protocolRemote_2 = 0
        self.assertEqual(times_protocolRemote_2, times_protocolRemote + 1)

        # Try to resend project, as no modification has been made number of
        # times should not change
        time.sleep(25)
        projectNotifier.notifyWorkflow()
        time.sleep(5)  # notifier runs in a thread so wait a bit
        results = json.loads(urlopen(urlProt).read().decode('utf-8'))
        objects = results["objects"]
        if len(objects) != 0:
            objects = results["objects"][0]
            times_protocolRemote_2 = objects["timesUsed"]
        else:
            times_protocolRemote_2 = 0
        self.assertEqual(times_protocolRemote_2, times_protocolRemote + 1)

        # Add new protocol and resend after waiting 30 = 25 +5  seconds
        prot2 = self.newProtocol(emprot.ProtStress, **kwargs)
        prot2.setObjLabel('stress2')
        self.proj.launchProtocol(prot2, wait=True)
        time.sleep(25)
        projectNotifier.notifyWorkflow()
        time.sleep(5)  # notifier runs in a thread so wait a bit
        results = json.loads(urlopen(urlProt).read().decode('utf-8'))
        objects = results["objects"]
        if len(objects) != 0:
            objects = results["objects"][0]
            times_protocolRemote_2 = objects["timesUsed"]
        else:
            times_protocolRemote_2 = 0

        self.assertEqual(times_protocolRemote_2, times_protocolRemote + 2)



# **************************************************************************
# *
# * Authors: David Herreros Calero    (dherreros@cnb.csic.es)
# *
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import numpy as np
import re

from pwem.convert.transformations import euler_matrix


class DraggablePoint:
    '''Class to create a dragabble point in Matplotlib using a scatter plot (in 3D)'''
    lock = None  # Only one can be animated at a time (probably useful in the future)
    def __init__(self, point, figure, axes, plot, M):
        self.point = point
        self.figure = figure
        self.axes = axes
        self.plot = plot
        self.M = M
        self.press = None
        self.background = None

    def getxyz(self, event):
        s = self.axes.format_coord(event.xdata, event.ydata)
        out = [float(re.sub(r'[^\x00-\x7F]+', '-', x.split('=')[1].strip())) for x in s.split(',')]
        return out

    def remove_projection_direction(self, coords, prev_point):
        prev_point = np.copy(prev_point)
        rot_matrix = euler_matrix(self.M[0], self.M[1], self.M[2], 'szyx')
        prev_point_rot = np.dot(rot_matrix, np.hstack([prev_point, 1]))
        new_point = np.dot(rot_matrix, np.hstack([coords, 1]))
        new_point[0] = prev_point_rot[0]
        coords = np.dot(np.linalg.inv(rot_matrix), new_point)
        return coords

    def connect(self):
        'connect to all the events we need'
        self.axes.disable_mouse_rotation()
        self.cidpress = self.figure.canvas.mpl_connect('button_press_event', self.on_press)
        self.cidrelease = self.figure.canvas.mpl_connect('button_release_event', self.on_release)
        self.cidmotion = self.figure.canvas.mpl_connect('motion_notify_event', self.on_motion)

    def on_press(self, event):
        if event.inaxes != self.axes:
            return
        if DraggablePoint.lock is not None:
            return
        coords = self.getxyz(event)
        coords = self.remove_projection_direction(coords, self.point)
        self.point = np.array((coords[0], coords[1], coords[2]))
        self.press = self.point, coords[0], coords[1], coords[2]
        DraggablePoint.lock = self

        # draw everything but the selected rectangle and store the pixel buffer
        canvas = self.figure.canvas
        axes = self.axes
        self.plot.set_animated(True)
        canvas.draw()
        self.background = canvas.copy_from_bbox(self.axes.bbox)

        # now redraw just the rectangle
        self.plot.set_offsets([coords[0], coords[1]])
        self.plot.set_3d_properties(coords[2], 'z')

        # and blit just the redrawn area
        self.plot.do_3d_projection(renderer=self.figure._cachedRenderer)
        canvas.restore_region(self.background)
        self.axes.draw_artist(self.plot)
        canvas.flush_events()
        canvas.blit(axes.bbox)

    def on_motion(self, event):
        if DraggablePoint.lock is not self:
            return
        if event.inaxes != self.axes:
            return
        self.point, xpress, ypress, zpress = self.press
        coords = self.getxyz(event)
        coords = self.remove_projection_direction(coords, self.point)
        dx = coords[0] - xpress
        dy = coords[1] - ypress
        dz = coords[2] - zpress
        self.point = np.array((self.point[0]+dx, self.point[1]+dy, self.point[2]+dz))

        canvas = self.figure.canvas
        axes = self.axes
        # restore the background region
        canvas.restore_region(self.background)

        # now redraw just the rectangle
        self.plot.set_offsets([self.point[0], self.point[1]])
        self.plot.set_3d_properties(self.point[2], 'z')

        # and blit just the redrawn area
        self.plot.do_3d_projection(renderer=self.figure._cachedRenderer)
        canvas.restore_region(self.background)
        self.axes.draw_artist(self.plot)
        canvas.flush_events()
        canvas.blit(axes.bbox)

    def on_release(self, event):
        'on release we reset the press data'
        if DraggablePoint.lock is not self:
            return

        self.press = None
        DraggablePoint.lock = None

        # turn off the rect animation property and reset the background
        self.plot.set_animated(False)
        self.background = None

        # redraw the full figure
        self.figure.canvas.draw()

    def disconnect(self):
        'disconnect all the stored connection ids'
        self.axes.mouse_init()
        self.figure.canvas.mpl_disconnect(self.cidpress)
        self.figure.canvas.mpl_disconnect(self.cidrelease)
        self.figure.canvas.mpl_disconnect(self.cidmotion)



# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     Roberto Marabini (roberto@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os

import pyworkflow.protocol.params as params
from pyworkflow import VERSION_3_0

from pwem.convert import Ccp4Header
from pwem.protocols import EMProtocol
import pwem.objects as emobj
import pyworkflow.utils as pwutils
from os.path import basename


class ProtOrigSampling(EMProtocol):
    """ Modify the origin and sampling values assigned to a 3D map
    """
    _label = 'assign Orig & Sampling'
    _program = ""
    _lastUpdateVersion = VERSION_3_0

    def __init__(self, **kwargs):
        EMProtocol.__init__(self, **kwargs)

    # --------------------------- DEFINE param functions ----------------------
    def _defineParams(self, form):
        form.addSection(label='Input')

        form.addParam('inVolume', params.PointerParam, label="Input Volume", important=True,
                      pointerClass='Volume',
                      help='A new object volume will be created with the assigned sampling rate and origin. '
                           'No new binary file will be created')
        form.addParam('copyFiles', params.BooleanParam,
                      label="Copy volume",
                      help="Option Yes:\nA new volume file will be copied "
                           "otherwise a link to the input volume is made\n"
                           "NOTE: if the copy option is selected the new sampling/origin"
                           " is stored in the header. Otherwise this information "
                           " is only stored in Scipion's database",
                      expertLevel=params.LEVEL_ADVANCED,
                      default=False)
        form.addParam('setSampling', params.BooleanParam,
                      label="Set sampling rate",
                      help="Option Yes:\nA new volume object will be created with "
                           "the given SamplingRate. "
                           "This sampling rate will NOT be set in the map file header.\n\n",
                      default=False)
        form.addParam('samplingRate', params.FloatParam,
                      condition='setSampling',
                      label=pwutils.Message.LABEL_SAMP_RATE)
        form.addParam('setOrigCoord', params.BooleanParam,
                      label="Set origin",
                      help="Option Yes:\nA new volume object will be created with "
                           "the given ORIGIN of coordinates. "
                           "This ORIGIN will NOT be set in the map file header.\n\n",
                      default=False)
        line = form.addLine("Coordinates (Å)",
                            help="A wizard will suggest you possible "
                                 "coordinates for the ORIGIN. In MRC volume "
                                 "files, the ORIGIN coordinates will be "
                                 "obtained from the file header.\n "
                                 "In case you prefer set your own ORIGIN "
                                 "coordinates, write them here. You have to "
                                 "provide the map center coordinates in "
                                 "Å (pixels x sampling).\n",
                                 condition='setOrigCoord')
        line.addParam('x', params.FloatParam,
                      label="x", default=0.,
                      help="offset along x axis (Å)")
        line.addParam('y', params.FloatParam,
                      label="  y", default=0.,
                      help="offset along y axis (Å)")
        line.addParam('z', params.FloatParam,
                      label="  z", default=0.,
                      help="offset along z axis (Å)")

    # --------------------------- INSERT steps functions ----------------------
    def _insertAllSteps(self):
        self._insertFunctionStep('assignStep')

    # --------------------------- STEPS functions -----------------------------        
        
    def assignStep(self):
        # Create a Volume object
        self.inVol = self.inVolume.get()
        self.outVol = emobj.Volume()
        self.outVol.copy(self.inVol)

        # Set sampling Rate (or copy from input)
        oldSampling = self.inVol.getSamplingRate()
        resample = True
        if self.setSampling.get():
            samplingRate = self.samplingRate.get()
        else:
            samplingRate = oldSampling
            resample = False

        self.outVol.setSamplingRate(samplingRate)

        # set Origin
        if self.setOrigCoord.get():
            origin = emobj.Transform()
            origin.setShifts(self.x.get(), self.y.get(), self.z.get())
        else:
            origin = self.inVol.getOrigin()
            if resample:
                x, y, z = origin.getShifts()
                factor = samplingRate / oldSampling
                origin.setShifts(x*factor, y*factor, z*factor)
            
        self.outVol.setOrigin(origin)

        # Files system stuff
        fileName = os.path.abspath(self.inVol.getFileName())
        fileName = fileName.replace(':mrc', '')

        # copy or link
        if self.copyFiles:
            imgBase = pwutils.replaceBaseExt(fileName, "mrc")
            imgDst = self._getExtraPath(imgBase)
            Ccp4Header.fixFile(fileName,imgDst, origin.getShifts(),
                               sampling=samplingRate)
        else:
            imgBase = basename(fileName)
            imgDst = self._getExtraPath(imgBase)
            pwutils.createAbsLink(fileName, imgDst)

        self.outVol.setLocation(imgDst)

        # save
        self._defineOutputs(outputVolume=self.outVol)
        self._defineSourceRelation(self.inVol, self.outVol)

    # --------------------------- INFO functions ------------------------------
    def _validate(self):
        message = []
        # message.append("You must set a path to export.")
        # Nothing to append
        return message

    def _summary(self):
        # volume copy or linked
        message = []
        if self.setSampling.get():
            message.append("New Sampling: %f\n" %
                           self.samplingRate)
        if self.setOrigCoord.get():
            message.append("New Origin: %f %f %f\n" %
                           (self.x, self.y, self.z))
        return message

    def _methods(self):
        return []

    # --------------------------- UTILS functions ---------------------------------

    def getFnPath(self, label='volume'):
        return os.path.join(self.filesPath.get(),
                            self._getFileName(label))



# **************************************************************************
# *
# * Authors:     R. Marabini (roberto@cnb.csic.es)
# *              Tomas Majtner (tmajtner@cnb.csic.es)   -- added particles
# *              Amaya Jimenez (ajimenez@cnb.csic.es) -- Mic for SetOfMics
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from collections import OrderedDict
from os.path import basename, splitext
import time
import random

import pyworkflow.protocol.params as params
from pyworkflow import VERSION_1_1
from pyworkflow.protocol.constants import STEPS_PARALLEL

from pwem.protocols import EMProtocol
import pwem.objects as emobj
from pwem import emlib


SET_OF_MOVIES = 0
SET_OF_MICROGRAPHS = 1
SET_OF_RANDOM_MICROGRAPHS = 2
SET_OF_PARTICLES = 3
SET_OF_COORDINATES = 4


class ProtCreateStreamData(EMProtocol):
    """ create  setofXXXX in streaming mode.
        micrograph -> read a micrograph in memory and writes it nDim times
        movie      -> read a movie in memory and writes it nDim times
        randomMicrographs -> creates a micrograph with random values
        and applies a random CTF
        particles  -> read nDim particles in memory and writes it in streaming
    """
    _label = "create stream data"
    _lastUpdateVersion = VERSION_1_1
    _singleImageFn = "singleImage.xmp"
    _magnification = 500000
    _voltage = 200
    _sphericalAberration = 2.0
    _amplitudeContrast = 0.07
    object = None

    def __init__(self, **kwargs):
        EMProtocol.__init__(self, **kwargs)
        self.dictObj = OrderedDict()
        self.stepsExecutionMode = STEPS_PARALLEL

    # -------------------------- DEFINE param functions ----------------------
    def _defineParams(self, form):
        form.addSection(label='Input')

        form.addParam('setof', params.EnumParam, default=0,
                      choices=['Movies', 'Micrographs', 'RandomMicrographs',
                               'Particles', 'Coordinates'],
                      label='set Of',
                      help='create set of')
        form.addParam('inputMovies', params.PointerParam, pointerClass='SetOfMovies',
                      condition="setof==%d" % SET_OF_MOVIES,
                      label="movie",
                      help='This movie will be copied "number of items" times')
        form.addParam('inputMics', params.PointerParam,
                      pointerClass='SetOfMicrographs',
                      condition="setof==%d" % SET_OF_MICROGRAPHS,
                      label="micrograph",
                      help='This micrograph will be copied "number of items"'
                           ' times')
        form.addParam('xDim', params.IntParam, default=1024,
                      condition="setof==%d" % SET_OF_RANDOM_MICROGRAPHS,
                      label="xdim",
                      help="X dim ")
        form.addParam('yDim', params.IntParam, default=1024,
                      condition="setof==%d" % SET_OF_RANDOM_MICROGRAPHS,
                      label="ydim",
                      help="Y dim ")
        form.addParam('inputParticles', params.PointerParam,
                      pointerClass='SetOfParticles',
                      condition="setof==%d" % SET_OF_PARTICLES,
                      label="SetOfParticles",
                      help='These particles will be written in streaming')
        form.addParam('inputCoordinates', params.PointerParam,
                      pointerClass='SetOfCoordinates',
                      condition="setof==%d" % SET_OF_COORDINATES,
                      label="SetOfCoordinates",
                      help='These Coordinates will be written in streaming')
        form.addParam('groups', params.IntParam, default=50,
                      condition="setof==%d" % SET_OF_PARTICLES,
                      label="groups",
                      help='How many items will be created every iteration')
        form.addParam('nDim', params.IntParam, default=10,
                      label="number of items",
                      help="setofXX size")
        form.addParam('samplingRate', params.FloatParam, default=4,
                      condition="setof!=%d and setof!=%d" % (
                          SET_OF_MICROGRAPHS, SET_OF_MOVIES),
                      label="samplingRate",
                      help="Sampling rate")
        form.addParam('creationInterval', params.IntParam, default=60,
                      label="Create Object each (sec)",
                      pointerClass='EMProtocol',
                      help="create one object each creationInterval seconds")
        form.addParam('extraRandomInterval', params.IntParam, default=0,
                      label='Extra random seconds interval', expertLevel=params.LEVEL_ADVANCED,
                      help='Each object will be generated in a random time uniformly picked from the interval defined by'
                           '[baseInterval, baseInterval+extraInterval]')
        form.addParam('delay', params.IntParam, default=0,
                      label="delay (sec)",
                      help="wait this seconds before creating stream data")

        form.addParallelSection(threads=1, mpi=1)

    # --------------------------- INSERT steps functions ---------------------
    def _insertAllSteps(self):
        self.counter = 0
        deps = []

        if self.delay.get() != 0:
            delayId = self._insertFunctionStep('delayStep')
            deps.append(delayId)

        step = None
        if self.setof == SET_OF_MOVIES:
            step = 'createStep'
        elif self.setof == SET_OF_MICROGRAPHS:
            step = 'createStep'
        elif self.setof == SET_OF_RANDOM_MICROGRAPHS:
            step = 'createRandomMicStep'
        elif self.setof == SET_OF_PARTICLES:
            step = 'createParticlesStep'
        elif self.setof == SET_OF_COORDINATES:
            step = 'createCoordinatesStep'
        else:
            raise Exception('Unknown data type')

        if self.setof == SET_OF_PARTICLES:
            self.nDims = int(min(self.nDim, len(self.inputParticles.get())))
            self.group = int(min(self.nDims, self.groups.get()))
            for mic in range(1, int((self.nDims / self.group) +
                             (self.nDims % self.group > 0) + 1)):
                self._insertFunctionStep(step, prerequisites=deps)

        elif self.setof == SET_OF_COORDINATES:
            self.inputMicrographs = self.inputCoordinates.get().getMicrographs()
            self.nDims = min(self.nDim.get(), len(self.inputMicrographs))
            for micIdx in range(1, self.nDims + 1):
                self._insertFunctionStep(step, micIdx, prerequisites=deps)

        else:
            for mic in range(1, self.nDim.get() + 1):
                self._insertFunctionStep(step, mic, prerequisites=deps)

    # -------------------------- STEPS functions ------------------------
    def delayStep(self):
        time.sleep(self.delay.get())

    def _checkNewItems(self, objSet):
        """ Check for already computed micrograph/movie and
        update the output set. """
        objDict = {}
        newObj = False
        for obj in objSet:
            objDict[obj.getFileName()] = True

        if objDict:
            if objSet.getSize():
                objSet.enableAppend()
                objSet.loadAllProperties()
        else:
            objSet.setStreamState(objSet.STREAM_OPEN)
            acquisition = emobj.Acquisition()
            if self.setof == SET_OF_MICROGRAPHS:
                acquisition.setMagnification(
                    self.inputMics.get().getAcquisition().getMagnification())
                acquisition.setVoltage(
                    self.inputMics.get().getAcquisition().getVoltage())
                acquisition.setSphericalAberration(
                    self.inputMics.get().getAcquisition().getSphericalAberration())
                acquisition.setAmplitudeContrast(
                    self.inputMics.get().getAcquisition().getAmplitudeContrast())
                objSet.setAcquisition(acquisition)
                objSet.setSamplingRate(self.inputMics.get().getSamplingRate())
            elif self.setof == SET_OF_MOVIES:
                acquisition.setMagnification(
                    self.inputMovies.get().getAcquisition().getMagnification())
                acquisition.setVoltage(
                    self.inputMovies.get().getAcquisition().getVoltage())
                acquisition.setSphericalAberration(
                    self.inputMovies.get().getAcquisition().getSphericalAberration())
                acquisition.setAmplitudeContrast(
                    self.inputMovies.get().getAcquisition().getAmplitudeContrast())
                objSet.setAcquisition(acquisition)
                objSet.setSamplingRate(
                    self.inputMovies.get().getSamplingRate())
            else:
                acquisition.setMagnification(self._magnification)
                acquisition.setVoltage(self._voltage)
                acquisition.setSphericalAberration(self._sphericalAberration)
                acquisition.setAmplitudeContrast(self._amplitudeContrast)
                objSet.setAcquisition(acquisition)
                if self.setof == SET_OF_PARTICLES:
                    objSet.setSamplingRate(
                        self.inputParticles.get().getSamplingRate())
                else:
                    objSet.setSamplingRate(self.samplingRate.get())

        if self.setof == SET_OF_MOVIES:
            obj = emobj.Movie()
        elif self.setof == SET_OF_MICROGRAPHS:
            obj = emobj.Micrograph()
        elif self.setof == SET_OF_RANDOM_MICROGRAPHS:
            obj = emobj.Micrograph()
        elif self.setof == SET_OF_PARTICLES:
            obj = emobj.Particle()
        else:
            raise Exception('Unknown data type')

        for k, v in self.dictObj.items():
            if k not in objDict:
                self.counter += 1
                obj.setFileName(k)
                if self.setof != SET_OF_PARTICLES:
                    obj.setMicName(basename(k))
                obj.setObjId(self.counter)
                objSet.append(obj)
                newObj = True

        return objSet, newObj  # why a dictionary, a boolean may be enough

    def _updateOutput(self, objSet):
        if self.setof == SET_OF_MOVIES:
            self._defineOutputs(outputMovies=objSet)
            self._defineTransformRelation(self.inputMovies, objSet)
        elif self.setof == SET_OF_MICROGRAPHS:
            self._defineOutputs(outputMicrographs=objSet)
            self._defineTransformRelation(self.inputMics, objSet)
        elif self.setof == SET_OF_RANDOM_MICROGRAPHS:
            self._defineOutputs(outputMicrographs=objSet)
        elif self.setof == SET_OF_PARTICLES:
            self._defineOutputs(outputParticles=objSet)
            self._defineTransformRelation(self.inputParticles, objSet)

    def _checkProcessedData(self):
        if self.setof == SET_OF_MOVIES:
            objSet = emobj.SetOfMovies(filename=self._getPath('movies.sqlite'))
        elif self.setof == SET_OF_MICROGRAPHS:
            objSet = emobj.SetOfMicrographs(filename=self._getPath('micrographs.sqlite'))
        elif self.setof == SET_OF_RANDOM_MICROGRAPHS:
            objSet = emobj.SetOfMicrographs(filename=self._getPath('micrographs.sqlite'))
        elif self.setof == SET_OF_PARTICLES:
            objSet = emobj.SetOfParticles(filename=self._getPath('particles.sqlite'))
        else:
            raise Exception('Unknown data type')

        newObjSet, newObj = self._checkNewItems(objSet)

        if self.setof == SET_OF_MOVIES:
            newObjSet.setFramesRange(self.inputMovies.get().getFramesRange())

        # check if end ....
        if self.setof != SET_OF_PARTICLES:
            self.nDims = self.nDim.get()
        endObjs = newObjSet.getSize() == self.nDims

        if newObj:
            if endObjs:
                newObjSet.setStreamState(newObjSet.STREAM_CLOSED)
            self._updateOutput(newObjSet)
        newObjSet.close()

    def createStep(self, counter):
        time.sleep(self.getTimeInterval())
        if not ProtCreateStreamData.object or self.setof == \
                SET_OF_MICROGRAPHS or self.setof == SET_OF_MOVIES:

            if self.setof == SET_OF_MOVIES:
                setDim = self.inputMovies.get().getSize()
                for idx, mov in enumerate(self.inputMovies.get()):
                    if idx == (counter - 1) % setDim:
                        newMov = mov.clone()
                        break
                _, ext = splitext(newMov.getFileName())
                ProtCreateStreamData.object = \
                    emlib.image.ImageHandler().read(newMov.getLocation())
                self.name = "movie"

            elif self.setof == SET_OF_MICROGRAPHS:
                setDim = self.inputMics.get().getSize()
                for idx, mic in enumerate(self.inputMics.get()):
                    if idx == (counter-1) % setDim:
                        newMic = mic.clone()
                        break
                _, ext = splitext(newMic.getFileName())
                ProtCreateStreamData.object = \
                    emlib.image.ImageHandler().read(newMic.getLocation())
                self.name = "micro"

        # save file
        destFn = self._getExtraPath("%s_%05d%s" % (self.name, counter, ext))
        ProtCreateStreamData.object.write(destFn)
        self.dictObj[destFn] = True
        self._checkProcessedData()

    def createParticlesStep(self):
        self.name = "particle"
        time.sleep(self.getTimeInterval())

        for idx, p in enumerate(self.inputParticles.get()):
            if ((idx > self.counter-1) and (idx < self.nDims) and
                    (idx <= self.counter-1 + self.group)):
                newP = p.clone()
                ProtCreateStreamData.object = \
                    emlib.image.ImageHandler().read(newP.getLocation())
                destFn = self._getExtraPath("%s_%05d" % (self.name, idx))
                ProtCreateStreamData.object.write(destFn)
                self.dictObj[destFn] = True
        self._checkProcessedData()

    def createCoordinatesStep(self, micIdx):
        self.name = "coordinate"
        time.sleep(self.getTimeInterval())
        inputCoordinates = self.inputCoordinates.get()
        micrographs = inputCoordinates.getMicrographs()
        if not hasattr(self, "outputCoordinates"):
            self.outputCoordinates = self._createSetOfCoordinates(micrographs)
            self.outputCoordinates.copyInfo(inputCoordinates)
            self.outputCoordinates.setBoxSize(inputCoordinates.getBoxSize())
            self.outputCoordinates.setStreamState(emobj.SetOfParticles.STREAM_OPEN)
            self._defineOutputs(outputCoordinates=self.outputCoordinates)
            self._defineSourceRelation(inputCoordinates, self.outputCoordinates)

        for idx, mic in enumerate(micrographs):
            if idx == micIdx - 1:
                newMic = mic.clone()
                for newCoord in self.inputCoordinates.get().iterCoordinates(newMic):
                    self.outputCoordinates.append(newCoord)
                if micIdx == self.nDims:
                    self.outputCoordinates.setStreamState(emobj.SetOfParticles.STREAM_CLOSED)
                self.outputCoordinates.write()
                self._store(self.outputCoordinates)
                break


    def createRandomMicStep(self, mic):
        from pwem import Domain
        time.sleep(self.getTimeInterval())
        getEnviron = Domain.importFromPlugin('xmipp3', 'Plugin',
                                             doRaise=True).getEnviron

        # create image
        img = emlib.Image()
        img.setDataType(emlib.DT_FLOAT)
        img.resize(self.xDim, self.yDim)
        img.initRandom(0., 1., emlib.XMIPP_RND_UNIFORM)
        baseFn = self._getExtraPath(self._singleImageFn)
        img.write(baseFn)

        md1 = emlib.MetaData()
        md1.setColumnFormat(False)
        idctf = md1.addObject()

        baseFnCtf = self._getTmpPath("ctf_%d.param" % mic)
        baseFnImageCTF = self._getExtraPath("imageCTF_%d.xmp" % mic)

        md1.setValue(emlib.MDL_CTF_SAMPLING_RATE, 1., idctf)
        md1.setValue(emlib.MDL_CTF_VOLTAGE, 200., idctf)
        defocus = 20000 + 10000 * random.random()
        udefocus = defocus + 1000 * random.random()
        vdefocus = defocus + 1000 * random.random()
        if udefocus < vdefocus:
            aux = vdefocus
            vdefocus = udefocus
            udefocus = aux
        md1.setValue(emlib.MDL_CTF_DEFOCUSU, udefocus, idctf)
        md1.setValue(emlib.MDL_CTF_DEFOCUSV, vdefocus, idctf)
        md1.setValue(emlib.MDL_CTF_DEFOCUS_ANGLE, 180.0 * random.random(),
                     idctf)
        md1.setValue(emlib.MDL_CTF_CS, 2., idctf)
        md1.setValue(emlib.MDL_CTF_Q0, 0.07, idctf)
        md1.setValue(emlib.MDL_CTF_K, 1., idctf)

        md1.write(baseFnCtf)

        # apply ctf
        args = " -i %s" % baseFn
        args += " -o %s" % baseFnImageCTF
        args += " -f ctf %s" % baseFnCtf
        args += " --sampling %f" % self.samplingRate
        self.runJob("xmipp_transform_filter", args, env=getEnviron())
        self.dictObj[baseFnImageCTF] = True
        self._checkProcessedData()

    # -------------------------- UTILS functions -----------------------------
    def getTimeInterval(self):
        baseInterval = self.creationInterval.get()
        interval = random.randint(baseInterval, baseInterval+self.extraRandomInterval.get())
        return interval


    # -------------------------- INFO functions ------------------------------
    def _validate(self):
        return []

    def _summary(self):
        return []

    def _methods(self):
        return []



# ***************************************************************************
# * Authors:     Roberto Marabini (roberto@cnb.csic.es)
# *
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# ***************************************************************************/
import os
from tempfile import mkstemp

import pyworkflow.tests as pwtests
import pwem.protocols as emprot

from xmipp3.protocols import XmippProtCreateMask3D


class TestOriginSampling(pwtests.BaseTest):
    @classmethod
    def setUpClass(cls):
        def generate(suffix="feat"):
            (fd, filename) = mkstemp(suffix=suffix)
            f = os.fdopen(fd, "w")
            fileContent = """# XMIPP_STAR_1 *
# Type of feature (sph, blo, gau, Cyl, dcy, cub, ell, con)(Required)
# The operation after adding the feature to the phantom (+/=) (Required)
# The feature density (Required)
# The feature center (Required)
# The vector for special parameters of each vector (Required)
# Sphere: [radius] Blob : [radius alpha m] Gaussian : [sigma]
# Cylinder : [xradius yradius height rot tilt psi]
# DCylinder : [radius height separation rot tilt psi]
# Cube : [xdim ydim zdim rot tilt psi]
# Ellipsoid : [xradius yradius zradius rot tilt psi]
# Cone : [radius height rot tilt psi]
data_block1
 _dimensions3D  '100 100 100'
 _phantomBGDensity  0.
 _scale  1.
data_block2
loop_
 _featureType
 _featureOperation
 _featureDensity
 _featureCenter
 _featureSpecificVector
blo = 1 '0 0 0' '50 10.4 2'
"""
            f.write(fileContent)
            f.close()
            return filename

        cls.inFileName = generate()
        pwtests.setupTestProject(cls)

    def test_assignOriginSampling(self):
        """ Create 3D mask
        """
        args = {'source': 2,  # mask from feature
                'featureFilePath': self.inFileName,
                'samplingRate': 1.1
                }

        prot = self.newProtocol(XmippProtCreateMask3D, **args)
        prot.setObjLabel('create mask 3d (feat file)')
        self.launchProtocol(prot)
        self.assertIsNotNone(prot.outputMask,
                             "There was a problem when importing a 3d mask.")

        # execute protocol that modify sampling/origin
        _sampling = 2.1
        _x = 11;
        _y = 22;
        _z = 33
        args = {'inVolume': prot.outputMask,
                'copyFiles': False,
                'setSampling': True,
                'samplingRate': _sampling,
                'setOrigCoord': True,
                'x': _x,
                'y': _y,
                'z': _z,
                }
        prot2 = self.newProtocol(emprot.ProtOrigSampling, **args)
        prot2.setObjLabel('set sampling and orig')
        self.launchProtocol(prot2)

        # read output and check new sampling
        vol = prot2.outputVolume
        sampling = vol.getSamplingRate()
        x, y, z = vol.getOrigin(force=True).getShifts()
        self.assertAlmostEqual(sampling, _sampling)
        self.assertAlmostEqual(x, _x)
        self.assertAlmostEqual(y, _y)
        self.assertAlmostEqual(z, _z)
        print("sampling=%f" % sampling)
        print("orig=%f %f %f" % (x, y, z))
        self.assertTrue(os.path.islink(vol.getFileName()), "%s is not a link" % vol.getFileName())

    def test_assignOrigin(self):
        """ Create 3D mask
        """
        args = {'source': 2,  # mask from feature
                'featureFilePath': self.inFileName,
                'samplingRate': 1.1
                }

        prot = self.newProtocol(XmippProtCreateMask3D, **args)
        prot.setObjLabel('create mask 3d (feat file)')
        self.launchProtocol(prot)
        self.assertIsNotNone(prot.outputMask,
                             "There was a problem when importing a 3d mask.")

        # execute protocol that modify sampling/origin
        _sampling = 2.1
        _x = 11;
        _y = 22;
        _z = 33
        args = {'inVolume': prot.outputMask,
                'copyFiles': False,
                'setSampling': False,
                'setOrigCoord': True,
                'x': _x,
                'y': _y,
                'z': _z,
                }
        prot2 = self.newProtocol(emprot.ProtOrigSampling, **args)
        prot2.setObjLabel('set orig')
        self.launchProtocol(prot2)

        # read output and check new sampling
        vol = prot2.outputVolume
        sampling = vol.getSamplingRate()
        x, y, z = vol.getOrigin(force=True).getShifts()
        self.assertAlmostEqual(x, _x)
        self.assertAlmostEqual(y, _y)
        self.assertAlmostEqual(z, _z)
        print("sampling=%f" % sampling)
        print("orig=%f %f %f" % (x, y, z))
        self.assertTrue(os.path.islink(vol.getFileName()))

    def test_assignSampling(self):
        """ Create 3D mask
        """
        args = {'source': 2,  # mask from feature
                'featureFilePath': self.inFileName,
                'samplingRate': 1.1
                }

        prot = self.newProtocol(XmippProtCreateMask3D, **args)
        prot.setObjLabel('create mask 3d (feat file)')
        self.launchProtocol(prot)
        self.assertIsNotNone(prot.outputMask,
                             "There was a problem when importing a 3d mask.")

        # execute protocol that modify sampling/origin
        _sampling = 2.1
        args = {'inVolume': prot.outputMask,
                'copyFiles': False,
                'setSampling': True,
                'samplingRate': _sampling,
                'setOrigCoord': False,
                }
        prot2 = self.newProtocol(emprot.ProtOrigSampling, **args)
        prot2.setObjLabel('set sampling')
        self.launchProtocol(prot2)

        # read output and check new sampling
        vol = prot2.outputVolume
        sampling = vol.getSamplingRate()
        x, y, z = vol.getOrigin(force=True).getShifts()
        self.assertAlmostEqual(sampling, _sampling)
        print("sampling=%f" % sampling)
        print("orig=%f %f %f" % (x, y, z))
        self.assertTrue(os.path.islink(vol.getFileName()))

    def test_assignOriginSamplingCopyFile(self):
        """ Create 3D mask
        """
        args = {'source': 2,  # mask from feature
                'featureFilePath': self.inFileName,
                'samplingRate': 1.1
                }

        prot = self.newProtocol(XmippProtCreateMask3D, **args)
        prot.setObjLabel('create mask 3d (feat file)')
        self.launchProtocol(prot)
        self.assertIsNotNone(prot.outputMask,
                             "There was a problem when importing a 3d mask.")

        # execute protocol that modify sampling/origin
        _sampling = 2.1
        _x = 11;
        _y = 22;
        _z = 33
        args = {'inVolume': prot.outputMask,
                'copyFiles': True,
                'setSampling': True,
                'samplingRate': _sampling,
                'setOrigCoord': True,
                'x': _x,
                'y': _y,
                'z': _z,
                }
        prot2 = self.newProtocol(emprot.ProtOrigSampling, **args)
        prot2.setObjLabel('set sampling, orig\n and copy file')
        self.launchProtocol(prot2)

        # read output and check new sampling
        vol = prot2.outputVolume
        sampling = vol.getSamplingRate()
        x, y, z = vol.getOrigin(force=True).getShifts()
        self.assertAlmostEqual(sampling, _sampling)
        self.assertAlmostEqual(x, _x)
        self.assertAlmostEqual(y, _y)
        self.assertAlmostEqual(z, _z)
        print("sampling=%f" % sampling)
        print("orig=%f %f %f" % (x, y, z))
        self.assertFalse(os.path.islink(vol.getFileName()))



# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *              Laura del Cano         (ldelcano@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This module contains protocols classes related to Random Conical Tilt.
"""

import sys
from os.path import basename
from glob import glob

import pyworkflow.utils as pwutils
import pyworkflow.protocol.params as params

import pwem.constants as emcts
from pwem import emlib
import pwem.objects as emobj

from .protocol_import import ProtImportFiles


class ProtImportMicrographsTiltPairs(ProtImportFiles):
    """Protocol to import untilted-tilted pairs of micrographs in the project"""
    
    _className = 'Micrograph'
    _label = 'import tilted micrographs'
    OUTPUT_NAME = "outputMicrographsTiltPair"
        
    # --------------------------- DEFINE param functions ----------------------
    
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addParam('patternUntilted', params.PathParam,
                      label=pwutils.Message.LABEL_PATTERNU,
                      help=pwutils.Message.TEXT_PATTERN)
        form.addParam('patternTilted', params.PathParam,
                      label=pwutils.Message.LABEL_PATTERNT,
                      help=pwutils.Message.TEXT_PATTERN)
        form.addParam('copyToProj', params.BooleanParam,
                      label=pwutils.Message.LABEL_COPYFILES, default=False)
        form.addParam('voltage', params.FloatParam, default=200,
                      label=pwutils.Message.LABEL_VOLTAGE, help=pwutils.Message.TEXT_VOLTAGE)
        form.addParam('sphericalAberration', params.FloatParam, default=2.26,
                      label=pwutils.Message.LABEL_SPH_ABERRATION,
                      help=pwutils.Message.TEXT_SPH_ABERRATION)
        form.addParam('ampContrast', params.FloatParam, default=0.1,
                      label=pwutils.Message.LABEL_AMPLITUDE,
                      help=pwutils.Message.TEXT_AMPLITUDE)
        form.addParam('samplingRateMode', params.EnumParam,
                      default=emcts.SAMPLING_FROM_IMAGE,
                      label=pwutils.Message.LABEL_SAMP_MODE,
                      help=pwutils.Message.TEXT_SAMP_MODE,
                      choices=[pwutils.Message.LABEL_SAMP_MODE_1,
                               pwutils.Message.LABEL_SAMP_MODE_2])
        form.addParam('samplingRate', params.FloatParam, default=1,
                      label=pwutils.Message.LABEL_SAMP_RATE,
                      help=pwutils.Message.TEXT_SAMP_RATE,
                      condition='samplingRateMode==%d' % emcts.SAMPLING_FROM_IMAGE)
        form.addParam('magnification', params.IntParam, default=50000,
                      label=pwutils.Message.LABEL_MAGNI_RATE,
                      help=pwutils.Message.TEXT_MAGNI_RATE,
                      condition='samplingRateMode==%d' % emcts.SAMPLING_FROM_SCANNER)
        form.addParam('scannedPixelSize', params.FloatParam, default=7.0,
                      label=pwutils.Message.LABEL_SCANNED,
                      condition='samplingRateMode==%d' % emcts.SAMPLING_FROM_SCANNER)

    # --------------------------- INSERT steps functions ----------------------
    
    def _insertAllSteps(self):
        self._insertFunctionStep('createTiltedPairsStep')
        
    # --------------------------- STEPS functions -----------------------------
    
    def importMicrographs(self, pattern, suffix, voltage, sphericalAberration, amplitudeContrast):
        """ Copy images matching the filename pattern
        Register other parameters.
        """
        filePaths = glob(pwutils.expandPattern(pattern))
        
        # imgSet = SetOfMicrographs(filename=self.micsPairsSqlite, prefix=suffix)
        imgSet = self._createSetOfMicrographs(suffix=suffix)
        acquisition = imgSet.getAcquisition()
        # Setting Acquisition properties
        acquisition.setVoltage(voltage)
        acquisition.setSphericalAberration(sphericalAberration)
        acquisition.setAmplitudeContrast(amplitudeContrast)
        
        # Call a function that should be implemented by each subclass
        self._setOtherPars(imgSet)
        
        outFiles = [imgSet.getFileName()]
        imgh = emlib.image.ImageHandler()
        img = imgSet.ITEM_TYPE()
        n = 1
        size = len(filePaths)
        
        filePaths.sort()
        
        for i, fn in enumerate(filePaths):
            # ext = os.path.splitext(basename(f))[1]
            dst = self._getExtraPath(basename(fn))
            if self.copyToProj:
                pwutils.copyFile(fn, dst)
            else:
                pwutils.createLink(fn, dst)
            
            if n > 1:
                for index in range(1, n+1):
                    img.cleanObjId()
                    img.setFileName(dst)
                    img.setIndex(index)
                    imgSet.append(img)
            else:
                img.cleanObjId()
                img.setFileName(dst)
                # Fill the micName if img is a Micrograph.
                self._fillMicName(img, fn, pattern)
                imgSet.append(img)
            outFiles.append(dst)
            
            sys.stdout.write("\rImported %d/%d" % (i+1, size))
            sys.stdout.flush()
            
        print("\n")
        
        imgSet.write()

        return imgSet
    
    def createTiltedPairsStep(self):
        args = {} 
        self.micsPairsSqlite = self._getPath('micrographs_pairs.sqlite')
        pwutils.cleanPath(self.micsPairsSqlite)  # Delete if exists
        
        micsTiltPair = emobj.MicrographsTiltPair(filename=self.micsPairsSqlite)
        micsU = self.importMicrographs(self.patternUntilted.get(), 'Untilted',
                                       self.voltage.get(),
                                       self.sphericalAberration.get(),
                                       self.ampContrast.get())
        micsT = self.importMicrographs(self.patternTilted.get(), 'Tilted',
                                       self.voltage.get(),
                                       self.sphericalAberration.get(),
                                       self.ampContrast.get())
        
        micsTiltPair.setUntilted(micsU)
        micsTiltPair.setTilted(micsT)
        
        for micU, micT in zip(micsU, micsT):
            micsTiltPair.append(emobj.TiltPair(micU, micT))
        
        args[self.OUTPUT_NAME] = micsTiltPair
        self._defineOutputs(**args)
    
    # --------------------------- INFO functions ------------------------------
    def _validate(self):
        errors = []
        if not self.patternUntilted.get() or not self.patternTilted.get():
            errors.append(pwutils.Message.ERROR_PATTERN_EMPTY)
        else:
            filePathsUntilted = glob(pwutils.expandPattern(self.patternUntilted.get()))
            filePathsTilted = glob(pwutils.expandPattern(self.patternTilted.get()))
        
            if len(filePathsUntilted) == 0 or len(filePathsTilted) == 0:
                errors.append(pwutils.Message.ERROR_PATTERN_FILES)

        return errors
    
    def _summary(self):
        summary = []
        output = self._getOutput()
        if not output:
            summary.append("Output not ready yet.") 
            if self.copyToProj:
                summary.append("*Warning*: Import step could take a long "
                               "time due to the images are copying in the project.")
        else:
            summary.append("Import of %d " % output.getTilted().getSize() + self._className + "s tilted from %s" % self.patternTilted.get())
            summary.append("Import of %d " % output.getUntilted().getSize() + self._className + "s untilted from %s" % self.patternUntilted.get())
            summary.append("Sampling rate : %0.2f Å/px" % self.samplingRate.get())
        
        return summary
    
    def _methods(self):
        methods = []
        output = self._getOutput()
        if output:
            methods.append("Imported %d micrographs tilted pairs %s " % (output.getTilted().getSize(), self.getObjectTag(output)))
            methods.append("with a sampling rate : %0.2f Å/px." % self.samplingRate.get())
            
        return methods
    
    # --------------------------- UTILS functions -----------------------------
    def getFiles(self):
        return getattr(self, self._getOutput().getFiles())
    
    def _getOutput(self):
        return getattr(self, self.OUTPUT_NAME, None)
        
    def _setOtherPars(self, micSet):
        micSet.getAcquisition().setMagnification(self.magnification.get())
        
        if self.samplingRateMode == emcts.SAMPLING_FROM_IMAGE:
            micSet.setSamplingRate(self.samplingRate.get())
        else:
            micSet.setScannedPixelSize(self.scannedPixelSize.get())
            
    def _fillMicName(self, img, filename, pattern):
        from pwem.objects import Micrograph
        if isinstance(img, Micrograph):
            filePaths = self.getMatchFiles(pattern=pattern)
            commPath = pwutils.commonPath(filePaths)
            micName = filename.replace(commPath + "/", "").replace("/", "_")
            img.setMicName(micName)



# **************************************************************************
# *
# * Authors:     Airen Zaldivar Peraza (azaldivar@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
In this module are protocol base classes related to EM.
There should be sub-classes in the different packages from
each EM-software package.
"""
from .protocol import *
from .protocol_micrographs import *
from .protocol_movies import *
from .protocol_align_movies import (ProtAlignMovies, ProtAverageFrames,
                                    createAlignmentPlot)
from .protocol_particles import *
from .protocol_particles_picking import *
from .protocol_2d import *
from .protocol_3d import *
from .protocol_sets import *
from .protocol_projection_edit import *
from .protocol_pdf_report import *
from .protocol_tiltpairs import *
from .protocol_ctf_assign import ProtCTFAssign
from .protocol_alignment_assign import ProtAlignmentAssign
from .protocol_alignment_invertHand import ProtAlignmentInvertHand
from .protocol_batch import *
from .protocol_extract_coordinates import ProtExtractCoords
from .protocol_create_stream_data import ProtCreateStreamData
from .parallel import ProtTestParallel
from .protocol_create_stream_data import SET_OF_RANDOM_MICROGRAPHS

from .protocol_import import *

from .protocol_export import *
from .protocol_tests import *

from .protocol_origin_sampling_volume import ProtOrigSampling
from .protocol_projection_edit import ProtProjectionEditor
from .protocol_set_editor import ProtSetEditor
from .protocol_set_filter import ProtSetFilter
from .protocol_movie_eraser import ProtMovieEraser
from .protocol_classes_selector import ProtClassesSelector
from .protocol_boxsize_parameters import ProtBoxSizeParameters
from .protocol_boxsize_checkpoint import ProtBoxSizeCheckpoint

from .protocol_mathematical_operator import ProtMathematicalOperator

from .protocol_tools import *




# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

# Expose many basic views
from .views import (DataView, ObjectView, MicrographsView, CtfView,
                    ClassesView, Classes3DView, CoordinatesObjectView,
                    ImageView, TableView)
from .viewer_base import EmProtocolViewer
from .plotter import EmPlotter
from .viewers_data import DataViewer

from .viewer_localres import LocalResolutionViewer, ChimeraAttributeViewer
from .viewer_vmd import Vmd, VmdView, VmdViewer
from .viewer_fsc import FscViewer
from .viewer_pdf import PDFReportViewer
from .viewer_chimera import (Chimera, ChimeraView, ChimeraViewer, ChimeraAngDist)
from .viewer_sequence import SequenceViewer
from .viewer_volumes import viewerProtImportVolumes
from .viewer_angular_dist import AngularDistributionViewer
from .viewers_data import BasicMDViewer

try:
    from .mdviewer import MDViewer
except Exception as e:
    pass

from .showj import *

if emConfig.CHIMERA_OLD_BINARY_PATH:
    from .viewer_chimera import ChimeraOldViewer



# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import sys
import os
import re

from os.path import exists, isdir
import time
from datetime import timedelta, datetime

import pyworkflow.utils as pwutils
import pyworkflow.protocol.params as params
from pwem import cleanFileName

from pwem.emlib.image import ImageHandler
from pwem.objects import Acquisition

from .base import ProtImportFiles


class ProtImportImages(ProtImportFiles):
    """ Common protocol to import a set of images into the project. """
    # The following class property should be set in each import subclass
    # for example, if set to SetOfParticles, this will the output classes
    # It is also assumed that a function with the name _createSetOfParticles
    # exists in the EMProtocol base class
    _outputClassName = 'None'  
    # If set to True, each binary file will be inspected to
    # see if it is a binary stack containing more items
    _checkStacks = True
        
    # --------------------------- DEFINE param functions ----------------------

    def _defineAcquisitionParams(self, form):
        """ Define acquisition parameters, it can be overwritten
        by subclasses to change what parameters to include.
        """
        group = form.addGroup('Acquisition info')
        group.addParam('haveDataBeenPhaseFlipped', params.BooleanParam,
                       default=False,
                       label='Have data been phase-flipped?',
                       help='Set this to Yes if the images have been ctf-phase '
                            'corrected.')
        group.addParam('acquisitionWizard', params.LabelParam, important=True,
                       condition=self._acquisitionWizardCondition(),
                       label='Use the wizard button to import acquisition.',
                       help='Depending on the import Format, the wizard\n'
                            'will try to import the acquisition values.\n'
                            'If not found, required ones should be provided.')
        group.addParam('voltage', params.FloatParam, default=300,
                       label=pwutils.Message.LABEL_VOLTAGE,
                       help=pwutils.Message.TEXT_VOLTAGE)
        group.addParam('sphericalAberration', params.FloatParam, default=2.7,
                       label=pwutils.Message.LABEL_SPH_ABERRATION,
                       help=pwutils.Message.TEXT_SPH_ABERRATION)
        group.addParam('amplitudeContrast', params.FloatParam, default=0.1,
                       label=pwutils.Message.LABEL_AMPLITUDE,
                       help=pwutils.Message.TEXT_AMPLITUDE)
        group.addParam('magnification', params.IntParam, default=50000,
                       label=pwutils.Message.LABEL_MAGNI_RATE,
                       help=pwutils.Message.TEXT_MAGNI_RATE)
        return group

    def _acquisitionWizardCondition(self):
        """ By default this wizard will appears only when we import from
        a format that is not from files.
        But movie-import also can have a wizard to read from FEI xml files. """
        return 'importFrom != %d' % self.IMPORT_FROM_FILES

    # --------------------------- INSERT functions ----------------------------
    def _insertAllSteps(self):

        # Only the import movies has property 'inputIndividualFrames'
        # so let's query in a non-intrusive manner
        inputIndividualFrames = getattr(self, 'inputIndividualFrames', False)

        if self.dataStreaming or inputIndividualFrames:
            funcName = 'importImagesStreamStep' 
        else:
            funcName = 'importImagesStep'
        
        self._insertFunctionStep(funcName, self.getPattern(),
                                 self.voltage.get(),
                                 self.sphericalAberration.get(),
                                 self.amplitudeContrast.get(),
                                 self.magnification.get())
        
    # --------------------------- STEPS functions -----------------------------
    def importImagesStep(self, pattern, voltage, sphericalAberration, 
                         amplitudeContrast, magnification):
        """ Copy images matching the filename pattern
        Register other parameters.
        """
        self.info("Using pattern: '%s'" % pattern)
        
        createSetFunc = getattr(self, '_create' + self._outputClassName)
        imgSet = createSetFunc()
        imgSet.setIsPhaseFlipped(self.haveDataBeenPhaseFlipped.get())
        acquisition = imgSet.getAcquisition()
        
        self.fillAcquisition(acquisition)
        
        # Call a function that should be implemented by each subclass
        self.setSamplingRate(imgSet)
        
        outFiles = [imgSet.getFileName()]
        imgh = ImageHandler()
        img = imgSet.ITEM_TYPE()
        img.setAcquisition(acquisition)
        n = 1
        copyOrLink = self.getCopyOrLink()
        alreadyWarned = False  # Use this flag to warn only once

        for i, (fileName, fileId) in enumerate(self.iterFiles()):
            if self.isBlacklisted(fileName):
                continue
            uniqueFn = self._getUniqueFileName(fileName)
            dst = self._getExtraPath(uniqueFn)
            if ' ' in dst:
                if not alreadyWarned:
                    self.warning('Warning: your file names have white spaces!')
                    self.warning('Removing white spaces from copies/symlinks.')
                    alreadyWarned = True
                dst = dst.replace(' ', '')
            copyOrLink(fileName, dst)
            # Handle special case of Imagic images, copying also .img or .hed
            self.handleImgHed(copyOrLink, fileName, dst)
            
            if self._checkStacks:
                _, _, _, n = imgh.getDimensions(dst)
                
            if n > 1:
                for index in range(1, n+1):
                    img.cleanObjId()
                    img.setMicId(fileId)
                    img.setFileName(dst)
                    img.setIndex(index)
                    self._addImageToSet(img, imgSet)
            else:
                img.setObjId(fileId)
                img.setLocation(dst)
                # Fill the micName if img is either Micrograph or Movie
                uniqueFn = uniqueFn.replace(' ', '')
                self._fillMicName(img, uniqueFn)
                self._addImageToSet(img, imgSet)

            outFiles.append(dst)
            
            sys.stdout.write("\rImported %d/%d\n" % (i+1, self.numberOfFiles))
            sys.stdout.flush()
            
        print("\n")
        
        args = {}
        outputSet = self._getOutputName()
        args[outputSet] = imgSet
        self._defineOutputs(**args)
        
        return outFiles

    def importImagesStreamStep(self, pattern, voltage, sphericalAberration,
                               amplitudeContrast, magnification):
        """ Copy images matching the filename pattern
        Register other parameters.
        """
        self.info("Using pattern: '%s'" % pattern)

        imgSet = self._getOutputSet() if self.isContinued() else None
        
        self.importedFiles = set()
        if imgSet is None:
            createSetFunc = getattr(self, '_create' + self._outputClassName)
            imgSet = createSetFunc()
        elif imgSet.getSize() > 0:  # in case of continue
            imgSet.loadAllProperties()
            self._fillImportedFiles(imgSet)
            imgSet.enableAppend()
        
        pointerExcludedMovs = getattr(self, 'moviesToExclude', None)
        if pointerExcludedMovs is not None:
            excludedMovs = pointerExcludedMovs.get()
            self._fillImportedFiles(excludedMovs)

        imgSet.setIsPhaseFlipped(self.haveDataBeenPhaseFlipped.get())
        acquisition = imgSet.getAcquisition()
        self.fillAcquisition(acquisition)
        # Call a function that should be implemented by each subclass
        self.setSamplingRate(imgSet)
        outFiles = [imgSet.getFileName()]
        imgh = ImageHandler()
        img = imgSet.ITEM_TYPE()
        img.setAcquisition(acquisition)
        n = 1
        copyOrLink = self.getCopyOrLink()
        outputName = self._getOutputName()
        alreadyWarned = False  # Use this flag to warn only once

        finished = False
        # this is only used when creating stacks from frame files
        self.createdStacks = set()

        i = 0
        lastDetectedChange = datetime.now()

        # Ignore the timeout variables if we are not really in streaming mode
        if self.dataStreaming:
            timeout = timedelta(seconds=self.timeout.get())
            fileTimeout = timedelta(seconds=self.fileTimeout.get())
        else:
            timeout = timedelta(seconds=5)
            fileTimeout = timedelta(seconds=5)

        while not finished:
            time.sleep(3)  # wait 3 seconds before check for new files
            someNew = False
            someAdded = False

            for fileName, uniqueFn, fileId in self.iterNewInputFiles():
                someNew = True
                if self.fileModified(fileName, fileTimeout):
                    continue
                
                dst = self._getExtraPath(uniqueFn)
                self.importedFiles.add(uniqueFn)
                dst, alreadyWarned = cleanFileName(dst, not alreadyWarned)

                copyOrLink(fileName, dst)

                self.debug('Importing file: %s' % fileName)
                self.debug("uniqueFn: %s" % uniqueFn)
                self.debug("dst Fn: %s" % dst)

                if self._checkStacks:
                    _, _, _, n = imgh.getDimensions(dst)

                someAdded = True
                self.debug('Appending file to DB...')
                if self.importedFiles:  # enable append after first append
                    imgSet.enableAppend()

                if n > 1:
                    for index in range(1, n+1):
                        img.cleanObjId()
                        img.setMicId(fileId)
                        img.setFileName(dst)
                        img.setIndex(index)
                        self._addImageToSet(img, imgSet)
                else:
                    img.setObjId(fileId)
                    img.setFileName(dst)
                    # Fill the micName if img is either a Micrograph or a Movie
                    uniqueFn = uniqueFn.replace(' ', '')
                    self.debug("FILENAME TO fillMicName: %s" % uniqueFn)
                    self._fillMicName(img, uniqueFn)
                    self._addImageToSet(img, imgSet)

                outFiles.append(dst)
                self.debug('After append. Files: %d' % len(outFiles))

            if someAdded:
                self.debug('Updating output...')
                self._updateOutputSet(outputName, imgSet,
                                      state=imgSet.STREAM_OPEN)
                self.debug('Update Done.')

            self.debug('Checking if finished...someNew: %s' % someNew)

            now = datetime.now()

            if not someNew:
                # If there are no new detected files, we should check the
                # inactivity time elapsed (from last event to now) and
                # if it is greater than the defined timeout, we conclude
                # the import and close the output set
                # Another option is to check if the protocol have some
                # special stop condition, this can be used to manually stop
                # some protocols such as import movies
                finished = (now - lastDetectedChange > timeout or
                            self.streamingHasFinished())
                self.debug("Checking if finished:")
                self.debug("   Now - Last Change: %s"
                           % pwutils.prettyDelta(now - lastDetectedChange))
                self.debug("Finished: %s" % finished)
            else:
                # If we have detected some files, we should update
                # the timestamp of the last event
                lastDetectedChange = now

        self._updateOutputSet(outputName, imgSet,
                              state=imgSet.STREAM_CLOSED)

        self._cleanUp()

        return outFiles

    @classmethod
    def validatePath(cls, pathstr):
        errors = []
        badChars = ['%', '#', ':']
        if any(bc in pathstr for bc in badChars):
            errors.append("File name can't contain these characters:")
            errors.append('  %s' % " ".join(badChars))
        return errors

    # --------------------------- INFO functions ------------------------------
    def _validateImages(self):
        errors = []
        ih = ImageHandler()

        for imgFn, _ in self.iterFiles():
            
            if isdir(imgFn):
                errors.append("Folders can not be selected.")
                errors.append('  %s' % imgFn)
            else:
                # Check if images are correct by reading the header of the
                # imported files:
                # Exceptions: 
                #  - Compressed movies (bz2 or tbz extensions)
                #  - Importing in streaming, since files may be incomplete
                #  - Bad characters in path [':' ,'%', '#']
                if (not self.dataStreaming and
                    not (imgFn.endswith('bz2') or
                         imgFn.endswith('tbz') or
                         imgFn.endswith('eer') or
                         ih.isImageFile(imgFn))):
                    if not errors:  # if empty add the first line
                        errors.append("Error reading the following images:")
                    errors.append('  %s' % imgFn)
                    errors += ProtImportImages.validatePath(imgFn)

            break  # validate just first image

        return errors

    def _validate(self):
        errors = ProtImportFiles._validate(self)

        # Check that files are proper EM images, only when importing from
        # files and not using streaming. In the later case we could
        # have partial files not completed.
        if (self.importFrom == self.IMPORT_FROM_FILES and
                not self.dataStreaming):
            errors += self._validateImages()

        return errors
        
    def _summary(self):
        summary = []
        outputSet = self._getOutputSet()
        if outputSet is None:
            summary.append("Output " + self._outputClassName + " not ready yet.") 
            if self.copyFiles:
                summary.append("*Warning*: You select to copy files into your "
                               "project.\n This will make another copy of "
                               "your data and may take more time to import.")
        else:
            summary.append("*%d* %s imported from %s" % (outputSet.getSize(),
                                                         self._getOutputItemName(),
                                                         self.getPattern()))
            summary.append("Is the data phase flipped : %s"
                           % outputSet.isPhaseFlipped())
            summary.append("Sampling rate : *%0.2f* Å/px"
                           % outputSet.getSamplingRate())
        
        return summary
    
    def _methods(self):
        methods = []
        outputSet = self._getOutputSet()
        if outputSet is not None:
            methods.append(u"*%d* %s were imported with a sampling rate of "
                           u"*%0.2f* Å/px (microscope voltage %d kV, "
                           u"magnification %dx). Output set is %s."
                           % (outputSet.getSize(), self._getOutputItemName(),
                              outputSet.getSamplingRate(),
                              round(self.voltage.get()),
                              round(self.magnification.get()),
                              self.getObjectTag(self._getOutputName())))
        return methods
    
    # --------------------------- UTILS functions -----------------------------
    def _cleanUp(self):
        """Empty method to override in child classes."""
        pass

    def getFiles(self):
        outputSet = self._getOutputSet()
        if outputSet is not None:
            return self._getOutputSet().getFiles()
        else:
            return []
    
    def _getOutputName(self):
        # We assume that the import output is always a 'SetOfSomething'
        return self._outputClassName.replace('SetOf', 'output')
    
    def _getOutputItemName(self):
        return self._outputClassName.replace('SetOf', '')
    
    def _getOutputSet(self):
        return getattr(self, self._getOutputName(), None)
    
    def fillAcquisition(self, acquisition):
        """ Fill the acquition object with protocol params. """
        acquisition.setVoltage(self.voltage.get())
        acquisition.setSphericalAberration(self.sphericalAberration.get())
        acquisition.setAmplitudeContrast(self.amplitudeContrast.get())
        acquisition.setMagnification(self.magnification.get())
        
    def getAcquisition(self):
        """ Build and fill an acquisition object. """
        acquisition = Acquisition()
        self.fillAcquisition(acquisition)
        
        return acquisition   
    
    def loadAcquisitionInfo(self):
        """ Return a proper acquistionInfo (dict)
        or an error message (str).
        """
        result = "Could not import acquisition info."
        # Check if the user selected to import from a given package
        # the property self.importFilePath should be set
        ci = self.getImportClass()

        if self.hasAttribute('importFilePath'):
            if not self.importFilePath:
                result = "Select import file first"
            elif not exists(self.importFilePath):
                result = ("*Import file doesn't exist.*\n"
                          "File:\n%s" % self.importFilePath)
            else:

                acq = ci.loadAcquisitionInfo()
                if acq is not None:
                    result = acq

        return result
    
    def _addImageToSet(self, img, imgSet):
        """ Add an image to a set, check if the first image to handle
         some special condition to read dimensions such as .txt or compressed
         movie files.
        """
        if imgSet.isEmpty():
            self._setupFirstImage(img, imgSet)
        imgSet.append(img)

    def _setupFirstImage(self, img, imgSet):
        pass

    def iterNewInputFiles(self):
        """ Iterate over input files that have not been imported.
        This function uses the self.importedFiles dict.
        """
        for fileName, fileId in self.iterFiles():
            # If file already imported or blacklisted,  skip it
            uniqueFn = self._getUniqueFileName(fileName)
            if (uniqueFn not in self.importedFiles) and (not self.isBlacklisted(fileName)):
                yield fileName, uniqueFn, fileId

    def _fillImportedFiles(self, imgSet):
        from pwem.objects import SetOfMicrographsBase
        if isinstance(imgSet, SetOfMicrographsBase):
            for img in imgSet:
                self.importedFiles.add(img.getMicName())

    def _fillMicName(self, img, uniqueFn):
        from pwem.objects import Micrograph
        if isinstance(img, Micrograph):
            img.setMicName(uniqueFn)
            
    def _getUniqueFileName(self, filename, filePaths=None):
        if filePaths is None:
            filePaths = [re.split(r'[$*#?]', self.getPattern())[0]]
        
        commPath = pwutils.commonPath(filePaths)
        return filename.replace(commPath + "/", "").replace("/", "_")

    def handleImgHed(self, copyOrLink, src, dst):
        """ Check the special case of Imagic files format
        composed by two files: .hed and .img.
        When copying or linking we need to take care
        of the other one.
        """
        if src.endswith('.hed'):
            src2 = src.replace('.hed', '.img')
            dst2 = dst.replace('.hed', '.img')
        elif src.endswith('.img'):
            src2 = src.replace('.img', '.hed')
            dst2 = dst.replace('.img', '.hed')
        else:
            src2 = None
            dst2 = None
        
        if src2 is not None and exists(src2):
            copyOrLink(src2, dst2)
    
    def _onNewFile(self, newFile):
        """ This method will be called we a new files is found in
        streaming mode. """
        pass

    def _createOutputSet(self):
        """ Create the output set that will be populated as more data is
        imported. """

    def processImportDict(self, importDict, importDir):
        """
        This function is used when we import a workflow from a json.
        If we need to include source data for reproducibility purposes,
        this function will make the necessary changes in the protocol dict
        to include the source data.
        Params:
            - importDict: import dictionary coming from the json
            - importDir: directory containing the json file
        """
        if importDict['filesPath']:
            if not os.path.exists(importDict['filesPath']):  # we might have a relative path
                absPath = os.path.join(importDir,
                                       importDict['filesPath'].split('/', 1)[1])
                if os.path.exists(absPath):
                    importDict['filesPath'] = absPath

        return importDict

    # --------------- Streaming special functions -----------------------
    def _getStopStreamingFilename(self):
        return self._getExtraPath("STOP_STREAMING.TXT")

    def getActions(self):
        """ This method will allow that the 'Stop import' action to appears
        in the GUI when the user right-click in the protocol import box.
        It will allow a user to manually stop the streaming.
        """
        # Only allow to stop if running and in streaming mode
        if self.dataStreaming and self.isRunning():
            return [('STOP STREAMING', self.stopImport)]
        else:
            return []

    def stopImport(self):
        """ Since the actual protocol that is running is in a different
        process that the one that this method will be invoked from the GUI,
        we will use a simple mechanism to place an special file to stop
        the streaming.
        """
        # Just place an special file into the run folder
        f = open(self._getStopStreamingFilename(), 'w')
        f.close()

    def streamingHasFinished(self):
        return os.path.exists(self._getStopStreamingFilename())



# **************************************************************************
# *
# * Authors:     Carlos Oscar S. Sorzano (coss@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import os
import glob

import pyworkflow as pw
import pyworkflow.protocol.params as params
import pyworkflow.utils as pwutils

from pwem.protocols import EMProtocol

PDFLATEX = "pdflatex"


class ProtPDFReport(EMProtocol):
    """ 
    Produce a pdf report from the files in a given directory.
    Supported file formats: *.tex, *.txt, *.jpg, *.png, *.pdf
    Files in the directory are sorted by name alphabetically,
    so if you want them to have the right order a possibility is to name them as
    0010-myText.txt
    0020-aFigure.png
    0030-anotherFigure.jpg
    0040-aPaper.pdf
    0050-anotherText.tex
    ...
    when these files are sorted, they will be sorted by the number in front.
    """    
    _label = 'pdf report'
    _lastUpdateVersion = pw.VERSION_1_1
    # --------------------------- DEFINE param functions ----------------------
    
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addParam('filesPath', params.PathParam,
                      label="Files directory",
                      help="Directory with the input files. \n"
                           "Check protocol help for more details.")
    
    # --------------------------- INSERT steps functions ----------------------
    
    def _insertAllSteps(self):
        self._insertFunctionStep('createReport')        

    # --------------------------- STEPS functions -----------------------------
    def createReport(self):
        fnTex = "report.tex"
        fhTex = open(self._getExtraPath(fnTex), "w")
        template = """
\\documentclass[12pt]{article}
\\usepackage{amsmath,amsthm,amssymb,amsfonts} 
\\usepackage{graphicx}
\\usepackage{pdfpages}
\\DeclareGraphicsExtensions{.pdf,.png,.jpg}
\\begin{document}
\\title{User Report}
\\author{Created by Scipion}
\\maketitle
"""
        fhTex.write(template)
        
        fnDir = self.filesPath.get()

        if not os.path.isdir(fnDir):
            fnDir = os.path.basename(fnDir)

        for fileName in sorted(glob.glob(os.path.join(fnDir, "*"))):
            fnDest = os.path.basename(fileName).lower()
            fnDest = fnDest.replace(" ", "_")
            fnDest = fnDest.replace(":", "_")
            fnDest = fnDest.replace(";", "_")
            pwutils.copyFile(fileName, self._getExtraPath(fnDest))

            if fnDest.endswith(".tex") or fnDest.endswith(".txt"):
                fhTex.write("\\input{%s}\n" % fnDest)
                fhTex.write("\n")
            elif fnDest.endswith(".png") or fnDest.endswith(".jpg"):
                fhTex.write("\\begin{center}\n")
                fhTex.write("\\includegraphics[width=14cm]{%s}\n" % fnDest)
                fhTex.write("\\end{center}\n")
                fhTex.write("\n")
            elif fnDest.endswith(".pdf"):
                fhTex.write("\\includepdf[pages=-]{%s}\n" % fnDest)
                fhTex.write("\\clearpage\n")
                fhTex.write("\n")
        
        template = """ 
\\end{document}
"""
        fhTex.write(template)
        fhTex.close()

        args = "-interaction=nonstopmode " + fnTex
        self.runJob(PDFLATEX, args, cwd=self._getExtraPath())

        fnPDF = self._getExtraPath("report.pdf")

        if os.path.exists(fnPDF):
            pwutils.moveFile(fnPDF, self._getPath("report.pdf"))
        else:
            raise Exception("PDF file was not produced.")

    # --------------------------- INFO functions ------------------------------
    def _summary(self):
        summary = []
        summary.append("Input directory: %s" % self.filesPath.get())
        summary.append("Output report: %s" % self._getPath("report.pdf"))
        return summary

    @classmethod
    def validateInstallation(cls):
        """ This function will be used to check if PDFto latex is available. """
        missingPaths = []

        from pyworkflow.utils import commandExists
        if not (commandExists(PDFLATEX)):
            missingPaths.append("%s not found in the system" % PDFLATEX)

        if missingPaths:
            return ["Missing variables:"] + missingPaths
        else:
            return []  # No errors



# **************************************************************************
# *
# * Authors:     Roberto Marabini (roberto@cnb.csic.es)
# *              Marta Martinez (mmmtnez@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
#
# cif to pdb conversion is based on  cif2pdb.py by Spencer Bliven
# <spencer.bliven@gmail.com>
#
# see http://biopython.org/DIST/docs/tutorial/Tutorial.html for a description
# on the object "structure" and other Bio.xxxx modules

from Bio.PDB.Dice import ChainSelector
from Bio.PDB.MMCIFParser import MMCIFParser
from Bio.PDB.PDBParser import PDBParser
from Bio.PDB import PDBIO, MMCIFIO
from Bio.PDB.mmcifio import mmcif_order
from Bio.PDB import Superimposer
from Bio.PDB.MMCIF2Dict import MMCIF2Dict
from Bio.PDB import Entity
from Bio.PDB import PDBList
from collections import OrderedDict
from Bio.PDB.Polypeptide import is_aa
from Bio.PDB.Polypeptide import three_to_one
from Bio.Seq import Seq
from pwem import Plugin, MAXIT_HOME
from pwem.convert.transformations import translation_from_matrix
import mmap
import re
import os
import numpy
import pyworkflow.utils as pwutils
import shutil

from pwem.objects.data import Alphabet



SECTION = '_scipion_attributes'
NAME, RECIP, SPEC, VALUE = SECTION + '.name', SECTION + '.recipient', SECTION + '.specifier', SECTION + '.value'

class OutOfChainsError(Exception):
    pass


class scipionMMCIFIO(MMCIFIO):

    # this class was needed before including maxit
    # as converter.
    # TODO: Delete it after a more extensive testing period
    # DATE mar feb 18 17:48:57 CET 2020

    def _save_dict_delete(self, out_file):
        # Form dictionary where key is first part of mmCIF key and value is list
        # of corresponding second parts
        key_lists = {}
        for key in self.dic:
            if key == "data_":
                data_val = self.dic[key]
            else:
                s = re.split(r"\.", key)
                if len(s) == 2:
                    if s[0] in key_lists:
                        key_lists[s[0]].append(s[1])
                    else:
                        key_lists[s[0]] = [s[1]]
                else:
                    raise ValueError("Invalid key in mmCIF dictionary: " + key)

        # Re-order lists if an order has been specified
        # Not all elements from the specified order are necessarily present
        for key, key_list in key_lists.items():
            # P3
            # for key, key_list in list(key_lists.items()):
            if key in mmcif_order:
                inds = []
                for i in key_list:
                    try:
                        inds.append(mmcif_order[key].index(i))
                    # Unrecognised key - add at end
                    except ValueError:
                        inds.append(len(mmcif_order[key]))
                key_lists[key] = [k for _, k in sorted(zip(inds, key_list))]

        # Write out top data_ line
        if data_val:
            out_file.write("data_" + data_val + "\n#\n")
            # ESCRIBIR POLYSEQTABLE
            out_file.write("""loop_
_entity_poly_seq.entity_id
_entity_poly_seq.num
_entity_poly_seq.mon_id
_entity_poly_seq.hetero\n#\n""")
            total_seqs = []
            counter_chain = 1
            for model in self.structure:
                for chain in model:
                    min_val = int(chain.get_unpacked_list()[0].id[1])
                    # max_val = min_val + len(chain.get_unpacked_list())
                    #                    not_are = range(1, min_val-1)
                    #                    for i in range(min_val)[:-1]:
                    #                        not_are.append(i + 1)
                    counter = 1
                    if min_val > 1:
                        for counter in range(1, min_val):
                            if len(chain.get_unpacked_list()[0].resname.strip()) == 3:
                                total_seqs.append((counter_chain, str(counter), "XAA", "n"))
                            elif len(chain.get_unpacked_list()[0].resname.strip()) == 2:
                                total_seqs.append((counter_chain, str(counter), "DX", "n"))
                            elif len(chain.get_unpacked_list()[0].resname.strip()) == 1:
                                total_seqs.append((counter_chain, str(counter), "X", "n"))
                            counter += 1
                    for residue in chain:
                        if len(chain.get_unpacked_list()[0].resname.strip()) == 3 and \
                                is_aa(residue.get_resname(), standard=True):  # aminoacids
                            total_seqs.append((counter_chain, residue.id[1], residue.get_resname(), "n"))
                        elif len(chain.get_unpacked_list()[0].resname.strip()) == 2 and \
                                residue.get_resname()[1] in ['A', 'C', 'G', 'T']:
                            total_seqs.append((counter_chain, counter, residue.get_resname(), "n"))
                        elif len(chain.get_unpacked_list()[0].resname.strip()) == 1 and \
                                residue.get_resname() in ['A', 'C', 'G', 'U']:
                            total_seqs.append((counter_chain, counter, residue.get_resname(), "n"))
                        counter += 1
                    counter_chain += 1
            for item in total_seqs:
                item = list(item)
                out_file.write("""%s %s %s  %s\n""" %
                               (str(item[0]), item[1], item[2], item[3]))

        for key, key_list in key_lists.items():
            #  for key, key_list in list(key_lists.items()):
            # Pick a sample mmCIF value, which can be a list or a single value
            sample_val = self.dic[key + "." + key_list[0]]
            n_vals = len(sample_val)
            # Check the mmCIF dictionary has consistent list sizes
            for i in key_list:
                val = self.dic[key + "." + i]
                if (isinstance(sample_val, list) and (isinstance(val, str) or len(val) != n_vals)) or (
                        isinstance(sample_val, str) and isinstance(val, list)):
                    raise ValueError("Inconsistent list sizes in mmCIF dictionary: " + key + "." + i)
            # If the value is a single value, write as key-value pairs
            if isinstance(sample_val, str):
                m = 0
                # Find the maximum key length
                for i in key_list:
                    if len(i) > m:
                        m = len(i)
                for i in key_list:
                    out_file.write(
                        "{k: <{width}}".format(k=key + "." + i, width=len(key) + m + 4) + self._format_mmcif_col(
                            self.dic[key + "." + i], len(self.dic[key + "." + i])) + "\n")
            # If the value is a list, write as keys then a value table
            elif isinstance(sample_val, list):
                out_file.write("loop_\n")
                col_widths = {}
                # Write keys and find max widths for each set of values
                for i in key_list:
                    out_file.write(key + "." + i + "\n")
                    col_widths[i] = 0
                    for val in self.dic[key + "." + i]:
                        len_val = len(val)
                        # If the value requires quoting it will add 2 characters
                        if self._requires_quote(val) and not self._requires_newline(val):
                            len_val += 2
                        if len_val > col_widths[i]:
                            col_widths[i] = len_val
                # Technically the max of the sum of the column widths is 2048

                # Write the values as rows
                for i in range(n_vals):
                    for col in key_list:
                        out_file.write(self._format_mmcif_col(self.dic[key + "." + col][i], col_widths[col] + 1))
                    out_file.write("\n")
            else:
                raise ValueError("Invalid type in mmCIF dictionary: " + str(type(sample_val)))
            out_file.write("#\n")


class AtomicStructHandler:
    """ Class that contain utilities to handle pdb/cif files"""
    PDB = 0
    CIF = 1

    def __init__(self, fileName=None, permissive=1):
        # The PERMISSIVE flag indicates that a number of common problems
        # associated with PDB files will be ignored
        self.permissive = permissive
        self.pdbParser = None
        self.cifParser = None
        self.ioPDB = None
        self.ioCIF = None
        self.structure = None
        self._readDone = False
        if fileName is not None:
            self.read(fileName)

    def readFromPDBDatabase(self, pdbID, dir=None, type='mmCif'):
        """
        Retrieve structure from PDB
        :param pdbID:
        :param dir: save structure in this directory
        :param type:  mmCif or pdb
        :return: filename with pdb file
        """
        if dir is None:
            dir = os.getcwd()
        pdbl = PDBList(pdb=dir)
        for counter in range(5):
            try:
                fileName = pdbl.retrieve_pdb_file(pdbID, pdir=dir, file_format=type)
            except Exception as error:
                print("Error: ", error)
                print("Retry connection", counter)
                if counter == 4:
                    raise
        return os.path.abspath(fileName)

    def checkLabelInFile(self, fileName, label):
        with open(fileName, "r") as f:
            s = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
            if s.find(label) != -1:
                return True
            else:
                return False

    def getStructure(self):
        """return structure information, model, chain,
           residues, atoms..."""
        return self.structure

    def read(self, fileName):
        """ Read and parse file."""
        # biopython assigns an ID to any read structure
        structure_id = os.path.basename(fileName)
        structure_id = structure_id[:4] if len(structure_id) > 4 else "1xxx"

        if fileName.endswith(".pdb") or fileName.endswith(".ent"):
            if self.pdbParser is None:
                self.pdbParser = PDBParser(PERMISSIVE=self.permissive)
            parser = self.pdbParser
            self.type = self.PDB
        else:
            if self.cifParser is None:
                self.cifParser = MMCIFParser()
            parser = self.cifParser
            self.type = self.CIF

        self.structure = parser.get_structure(structure_id, fileName)
        self._readDone = True

    def checkRead(self):
        if self._readDone:
            return True
        else:
            print("you must read the pdb file first")
            exit(0)

    def getModelsChains(self):
        """
        given an atomic structure returns two dictionaries:
            (1) for all models and respective chains (chainID and length of residues)
            (2) for each chain list of residues
        """
        self.checkRead()
        listOfChains = OrderedDict()
        listOfResidues = OrderedDict()

        for model in self.structure:
            chainDicLength = OrderedDict()
            chainDicFirstResidue = OrderedDict()
            for chain in model:
                if len(chain.get_unpacked_list()[0].resname.strip()) == 1:  # RNA
                    seq = list()
                    seq_number = list()
                    for residue in chain:
                        if residue.get_resname() in ['A', 'C', 'G', 'U']:
                            seq.append(residue.get_resname())
                        else:
                            seq.append("X")
                        seq_number.append((residue.get_id()[1], residue.get_resname()))
                elif len(chain.get_unpacked_list()[0].resname.strip()) == 2:  # DNA
                    seq = list()
                    seq_number = list()
                    for residue in chain:
                        if residue.get_resname()[1] in ['A', 'C', 'G', 'T']:
                            seq.append(residue.get_resname()[1])
                        else:
                            seq.append("X")
                        seq_number.append((residue.get_id()[1], residue.get_resname()))
                elif len(chain.get_unpacked_list()[0].resname.strip()) == 3:  # Protein
                    seq = list()
                    seq_number = list()
                    counter = 0
                    for residue in chain:
                        if is_aa(residue.get_resname(), standard=True):  # aminoacids
                            seq.append(three_to_one(residue.get_resname()))
                            counter += 1
                        else:
                            seq.append("X")
                        seq_number.append((residue.get_id()[1], residue.get_resname()))
                    if counter == 0:  # HETAM
                        for residue in chain:
                            seq.append(residue.get_resname())
                while seq[-1] == "X":
                    del seq[-1]
                while seq[0] == "X":
                    del seq[0]
                chainDicLength[chain.id] = len(seq)
                chainDicFirstResidue[chain.id] = seq_number
            listOfChains[model.id] = chainDicLength
            listOfResidues[model.id] = chainDicFirstResidue

        return listOfChains, listOfResidues

    def getSequenceFromChain(self, modelID, chainID, returnAlphabet=False):
        self.checkRead()
        seq = list()
        for model in self.structure:
            if model.id == modelID:
                for chain in model:
                    if str(chain.id) == chainID:
                        if len(chain.get_unpacked_list()[0].resname) == 1:
                            print("Your sequence is a nucleotide sequence ("
                                  "RNA)\n")
                            alphabet = Alphabet.UNAMBIGOUS_RNA_ALPHABET
                            for residue in chain:
                                # Check if the residue belongs to the
                                # standard RNA and add those residues to the
                                # seq
                                if residue.get_resname() in ['A', 'C',
                                                             'G', 'U']:
                                    seq.append(residue.get_resname())
                                else:
                                    seq.append("X")
                        elif len(chain.get_unpacked_list()[0].resname) == 2:
                            print("Your sequence is a nucleotide sequence ("
                                  "DNA)\n")
                            alphabet = Alphabet.UNAMBIGOUS_DNA_ALPHABET
                            for residue in chain:
                                # Check if the residue belongs to the
                                # standard DNA and add those residues to the
                                # seq
                                if residue.get_resname()[1] in ['A', 'C',
                                                                'G', 'T']:
                                    seq.append(residue.get_resname()[1])
                                else:
                                    seq.append("X")
                        elif len(chain.get_unpacked_list()[0].resname) == 3:
                            counter = 0
                            for residue in chain:
                                if is_aa(residue.get_resname(), standard=True):
                                    alphabet = Alphabet.EXTENDED_PROTEIN_ALPHABET
                                    # The test checks if the amino acid
                                    # is one of the 20 standard amino acids
                                    # Some proteins have "UNK" or "XXX", or other symbols
                                    # for missing or unknown residues
                                    seq.append(three_to_one(residue.get_resname()))
                                    counter += 1
                                else:
                                    seq.append("X")
                            if counter != 0:  # aminoacids
                                print("Your sequence is an aminoacid sequence")
                            else:  # HETAM
                                print("Your sequence is a HETAM sequence")
                                for residue in chain:
                                    seq.append(residue.get_resname())
                        while seq[-1] == "X":
                            del seq[-1]
                        while seq[0] == "X":
                            del seq[0]
                        # return Seq(str(''.join(seq)), alphabet=alphabet)
                        if returnAlphabet:
                            return Seq(str(''.join(seq))), alphabet
                        else:
                            return Seq(str(''.join(seq)))

    def getFullID(self, model_id='0', chain_id=None):
        """
        assign a label to a sequence obtained from a PDB file
        :parameter
        model_id
        chain_id
        :return: string with label
        """
        self.checkRead()  # cehck we have read the structure
        label = "%s" % self.structure.get_id()  # PDB_ID
        # check how many model are in sequence if > 1 add to id
        models = self.getModelsChains()
        if len(models) > 1:
            label = label + "__%s" % str(model_id)
        # if chain_id is None do not add it to te name
        if chain_id is not None:
            label = label + "_%s" % str(chain_id)
        return label

    def getStructureBFactorValues(self):
        """Using an atomic structure as input, this method returns two list:
        * List 1: B-factor field values for each atom of the structure.
        * List 2 is a list of lists. For each residue of the structure a list with
        the model, chain, order number of residue, name of residue and average of
        B-factor field values (atoms) is returned. An example
        of each list item is:
        [0, 'A', 103, 'PHE', 96.01]"""
        self.checkRead()
        listOfBFactors = []
        listOfResiduesBFactors = []
        for model in self.structure:
            for chain in model:
                for residue in chain:
                    residue_list = []
                    residue_list.append(model.id)
                    residue_list.append(chain.id)
                    residue_list.append(residue.id[1])
                    residue_list.append(residue.get_resname())
                    sum_list = 0.00
                    residue_atom_b_factor_list = []
                    for atom in residue:
                        listOfBFactors.append(atom.get_bfactor())
                        residue_atom_b_factor_list.append(atom.get_bfactor())
                    for item in residue_atom_b_factor_list:
                        sum_list += item
                    avg = sum_list / len(residue_atom_b_factor_list)
                    residue_list.append(round(avg, 2))
                    listOfResiduesBFactors.append(residue_list)
        return listOfBFactors, listOfResiduesBFactors

    def readLowLevel(self, fileName):
        """ Return a dictionary with all mmcif fields. you should parse them
            Example: get the list of the y coordinates of all atoms
              dict = readLowLevel("kk.pdb")
              y_list = dict['_atom_site.Cartn_y']
        """

        if fileName.endswith(".pdb"):
            print("Low level access to PDB is not implemented")
        else:
            dict = MMCIF2Dict(fileName)
        return dict

    def _write(self, fileName):
        """ Do not use this function use toPDB or toCIF, they take care of some
        compatibiity issues"""
        if fileName.endswith(".pdb") or fileName.endswith(".ent"):
            if self.ioPDB is None:
                self.ioPDB = PDBIO()
            io = self.ioPDB
        else:
            if self.ioCIF is None:
                self.ioCIF = scipionMMCIFIO()
                # self.ioCIF = MMCIFIO()
            io = self.ioCIF
        io.set_structure(self.structure)
        io.save(fileName)

    def _writeLowLevel(self, fileName, dict):
        """ write a dictionary as cif file
        """

        if fileName.endswith(".pdb"):
            print("Low level access to PDB is not implemented")
        else:
            if self.ioCIF is None:
                self.ioCIF = MMCIFIO()
            io = self.ioCIF
        io.set_dict(dict)
        io.save(fileName)

    def _intToChain(self, i, base=62):
        """
        int_to_chain(int,int) -> str
        Converts a positive integer to a chain ID. Chain IDs include uppercase
        characters, numbers, and optionally lowercase letters.
        i = a positive integer to convert
        base = the alphabet size to include. Typically 36 or 62.
        """
        if i < 0:
            raise ValueError("positive integers only")
        if base < 0 or 62 < base:
            raise ValueError("Invalid base")

        quot = int(i) // base
        rem = i % base
        if rem < 26:
            letter = chr(ord("A") + rem)
        elif rem < 36:
            letter = str(rem - 26)
        else:
            letter = chr(ord("a") + rem - 36)
        if quot == 0:
            return letter
        else:
            return self._intToChain(quot - 1, base) + letter

    def renameChains(self, structure):
        """Renames chains to be one-letter chains

        Existing one-letter chains will be kept.
        Multi-letter chains will be truncated
        or renamed to the next available letter of the alphabet.

        If more than 62 chains are present in the structure,
        raises an OutOfChainsError

        Returns a map between new and old chain IDs, as well as modifying
        the input structure

        """
        next_chain = 0  #
        # single-letters stay the same
        chainmap = {c.id: c.id
                    for c in structure.get_chains() if len(c.id) == 1}
        for o in structure.get_chains():
            if len(o.id) != 1:
                if o.id[0] not in chainmap:
                    chainmap[o.id[0]] = o.id
                    o.id = o.id[0]
                else:
                    c = self._intToChain(next_chain)
                    while c in chainmap:
                        next_chain += 1
                        c = self._intToChain(next_chain)
                        if next_chain >= 62:
                            raise OutOfChainsError()
                    chainmap[c] = o.id
                    o.id = c
        return chainmap

    def write(self, fileName):
        if fileName.endswith(".pdb") or fileName.endswith(".ent"):
            self.writeAsPdb(fileName)
        else:
            self.writeAsCif(fileName)

    def writeAsPdb(self, pdbFile):
        """ Save structure as PDB. Be aware that this is not a lossless conversion
        Returns False is conversion is not possible. True otherwise
        """
        # check input is not PDB
        if self.type == self.PDB:
            pass
        else:
            # rename long chains
            try:
                chainmap = self.renameChains(self.structure)
            except OutOfChainsError:
                print("Too many chains to represent in PDB format")
                return False

            for new, old in chainmap.items():
                # for new, old in list(chainmap.items()):
                if new != old:
                    print("Renaming chain {0} to {1}".format(old, new))

        self._write(pdbFile)

        return True

    def writeAsCif(self, cifFile):
        """ Save structure as CIF.
            Be aware that this is not a lossless conversion
        """
        self._write(cifFile)

    def getBoundingBox(self, expand=3):
        """Get bounding box (angstroms) for atom struct.
         Only alpha carbons are taken into account.
         parameter: expand.- make box larger addind this factor
         """
        # Use the first model
        ref_model = self.getStructure()[0]
        # init bounding box volues
        xmin = 100000; ymin = xmin; zmin = xmin
        xmax = -100000; ymax = xmax; zmax = xmax
        # iterate for all aminoacids
        for chain in ref_model:
            for residue in chain:
                (x,y,z) = residue["CA"].get_coord()
                if x < xmin: xmin = x
                if x > xmax: xmax = x

                if y < ymin: ymin = y
                if y > ymax: ymax = y

                if z < zmin: zmin = z
                if z > zmax: zmax = z
            # DEBUG
            # with open("/tmp/kk.bild", "w") as file:
            #     file.write(".transparency 0.8\n"
            #                ".color red\n"
            #                ".box %f %f %f %f %f %f\n" %
            #                (xmin - expand , ymin - expand, zmin - expand,
            #                xmax + expand, ymax + expand, zmax + expand))
        return [[xmin - expand, ymin - expand, zmin - expand],
                [xmax + expand, ymax + expand, zmax + expand]]


    def getTransformMatrix(self, atomStructFn, startId=-1, endId=-1, outFn=None):
        """find matrix that Superimposes two atom structures.
        this matrix moves atomStructFn to self """
        if endId == -1:
            endId = 10000000
        # load second atom structure
        aSH = AtomicStructHandler()
        aSH.read(atomStructFn)

        # Use the first model in the atom struct for alignment
        ref_model = self.getStructure()[0]
        sample_model = aSH.getStructure()[0]

        # Make a list of the atoms (in the structures) you wish to align.
        # In this case we use CA atoms whose index is in the
        # specified range (starId, endId)
        ref_atoms = []
        sample_atoms = []

        # Now get a list with CA atoms (alpha-carbon)
        # I assume that both atom structs have the same number of chains
        # and in the same order.
        for ref_chain, sample_chain in zip(ref_model, sample_model):
            # create a set with the id of all residues
            # for the current chain
            ref_set_id = set(res.get_id()[1] for res in ref_chain)
            sample_set_id = set(res.get_id()[1] for res in sample_chain)
            # keep the intersection as an ordered list (sets are not ordered)
            ref_set_id.intersection_update(sample_set_id)
            ref_list_id = sorted(ref_set_id)
            # delete AA smaller or large than the predefined values
            ref_list_id[:] = [x for x in ref_list_id
                              if x >= startId and x <= endId]

            # Iiterate through all residues in each chain a store CA atoms
            for id in ref_list_id:
                if id in ref_chain and id in sample_chain:
                    ref_atoms.append(ref_chain[id]['CA'])
                    sample_atoms.append(sample_chain[id]['CA'])
        # Now we initiate the superimposer:
        super_imposer = Superimposer()
        super_imposer.set_atoms(ref_atoms, sample_atoms)
        # super_imposer.apply(sample_model.get_atoms())
        (rot, trans) = super_imposer.rotran
        # DEBUG, uncomment next two lines to see
        # transformation applied to atomStructFn
        if outFn:
            aSH.getStructure().transform(rot, trans)
            aSH.write(outFn)

        # convert 3x3 rotation matrix to homogeneous matrix
        rot = numpy.transpose(rot)  # scipion and biopython use
                                    # different conventions
        tmp = numpy.r_[rot, numpy.zeros((1, 3))]
        mat = numpy.c_[tmp, numpy.array(
            [[trans[0]], [trans[1]], [trans[2]], [1]])]
        return mat, super_imposer.rms

    def centerOfMass(self, geometric=False):
        """
        Returns gravity [default] or geometric center of mass of an Entity
        (anything with a get_atoms function in biopython.
        Geometric assumes all masses are equal (geometric=True)
        """
        entity = self.structure
        # Structure, Model, Chain, Residue
        if isinstance(entity, Entity.Entity):
            atom_list = entity.get_atoms()
        # List of Atoms
        elif hasattr(entity, '__iter__') and \
                [x for x in entity if x.level == 'A']:
            atom_list = entity
        else:  # Some other weirdo object
            raise ValueError("Center of Mass can only be calculated "
                             "from the following objects:\n"
                             "Structure, Model, Chain, Residue, "
                             "list of Atoms.")

        masses = []
        positions = [[], [], []]

        for atom in atom_list:
            masses.append(atom.mass)

            for i, coord in enumerate(atom.coord.tolist()):
                positions[i].append(coord)

        # If there is a single atom with undefined mass complain loudly.
        if 'ukn' in set(masses) and not geometric:
            raise ValueError("Some Atoms don't have an element assigned.\n"
                             "Try adding them manually or calculate the "
                             "geometrical center of mass instead.")

        if geometric:
            return [sum(coord_list) / len(masses) for coord_list in positions]
        else:
            w_pos = [[], [], []]
            for atom_index, atom_mass in enumerate(masses):
                w_pos[0].append(positions[0][atom_index] * atom_mass)
                w_pos[1].append(positions[1][atom_index] * atom_mass)
                w_pos[2].append(positions[2][atom_index] * atom_mass)

            return [sum(coord_list) / sum(masses) for coord_list in w_pos]

    def transform(self, transformation_matrix, sampling=1.):
        """ Geometrical transformation of a PDB structure

        :param entity: PDB biopython structure
        :param transformation_matrix -> 4x4 scipion matrix
        :param sampling: scipion transform matrix is applied to voxels so
              length must be multiplied by samplingRate

        internal variables:
             rotation matrix -> numpy.array(
                      [[ 0.5, -0.809017,  0.309017],
                       [ 0.809017,  0.30917,  -0.5     ],
                       [ 0.309017,  0.5,       0.809017]])
             translation: translation vector -> numpy.array([1., 0., 0.], 'd')
        :return: no return, new data overwrites entity
        """
        # bioPhython and Scipion conventions do not match
        rotation_matrix = numpy.transpose(transformation_matrix[:3, :3])
        # from geometry get euler angles and recreate matrix
        translation = translation_from_matrix(transformation_matrix)
        translation = [x * sampling for x in translation]
        self.structure.transform(rotation_matrix, translation)

    def _renameChainsIfNeed(self, struct2):
        """Rename chain, we assume that there is a single model per structure"""
        repeated = False

        def RepresentsInt(s):
            try:
                int(s)
                return True
            except ValueError:
                return False

        import uuid
        chainIDs1 = [chain.id for chain in self.structure.get_chains()]
        for chain in struct2.get_chains():
            counter = 2
            if chain.id in chainIDs1:
                repeated = True
                cId = chain.id
                l = len(cId)
                if l == 1:
                    while True:
                        try:
                            chain.id = "%s%03d" % (cId, counter)
                            break
                        except ValueError:
                            counter +=1
                            if counter > 1000:
                                raise ValueError('Error in _renameChainsIfNeed.')
                elif RepresentsInt(cId[1:]):  # try to fit a number and increase it by one
                    chain.id = "%s%03d" % (cId[0], int(cId[1:]) + 1)
                else:  # generate a 4 byte random string
                    chain.id = uuid.uuid4().hex[:4]
        if repeated:
            self._renameChainsIfNeed(struct2)

    def addStruct(self, secondPDBfileName, outPDBfileName=None, useModel=False):
        """ Join the second structure to the first one.
            If cheon numes are the same rename them.
            if outPDBfileName id provided then new
            struct is saved to a file"""
        # read new structure
        if outPDBfileName is not None:
            pdbID = (os.path.splitext(os.path.basename(outPDBfileName))[0])[:4]
        else:
            pdbID = (os.path.splitext(os.path.basename(secondPDBfileName))[0])[:4]

        if secondPDBfileName.endswith(".pdb") or secondPDBfileName.endswith(".ent"):
            parser = PDBParser(PERMISSIVE=self.permissive)
        else:
            parser = MMCIFParser()

        struct2 = parser.get_structure(pdbID, secondPDBfileName)

        if useModel:
            modelNumber = 0
            modelID = 0
            # model.id = model.serial_num = len(self.structure)?  # not sure this
            # is valid always
            for model in self.structure:
                pass
            modelNumber = model.serial_num
            modelID = model.id
            for model in struct2:
                modelNumber += 1
                modelID += 1
                model.detach_parent()
                model.serial_num = modelNumber
                model.id = modelID
                self.structure.add(model)
        else:
            self._renameChainsIfNeed(struct2)

            for model in struct2:
                for chain in model:
                    chain.detach_parent()
                    self.structure[0].add(chain)

        # create new output file
        if outPDBfileName is not None:
            self.write(outPDBfileName)

    def extractChain(self, chainID, start=0, end=-1,
                     modelID='0', filename="output.mmcif"):
        """Code for chopping  a structure.
        This module is used internally by the Bio.PDB.extract() function.
        """
        sel = ChainSelector(chain_id=chainID, start=start,
                            end=end, model_id=int(modelID))
        io = scipionMMCIFIO()
        io.set_structure(self.structure)
        io.save(filename, sel)

    def renameChain(self, chainID, newChainName, modelID='0',
                    filename="output.mmcif"):
        self.structure[modelID][chainID].id = newChainName
        self.write(filename)

    def renumberChain(self, chainID, offset=0, modelID='0',
                      filename="output.mmcif"):
        # get chain object
        chain = self.structure[modelID][chainID]
        # remove chain from model
        self.structure[modelID].detach_child(chainID)
        from Bio.PDB.Chain import Chain
        # create new chain
        newChain = Chain(chainID)
        for residue in chain:
            # remove residue, otherwise we cannot renumber it
            residue.detach_parent()
            rId = residue.id
            res_id = list(rId)
            res_id[1] = res_id[1] + offset
            if res_id[1] < 0:
                raise ValueError('Residue number cant be <= 0')
            residue.id = tuple(res_id)
            newChain.add(residue)
        self.structure[modelID].add(newChain)
        self.write(filename)


def cifToPdb(fnCif, fnPdb):
    h = AtomicStructHandler()
    h.read(fnCif)
    h.writeAsPdb(fnPdb)


def pdbToCif(fnPdb, fnCif):
    h = AtomicStructHandler()
    h.read(fnPdb)
    h.writeAsCif(fnCif)


def toPdb(inFileName, outPDBFile):
    if inFileName.endswith(".pdb") or inFileName.endswith(".ent"):
        return inFileName
    elif inFileName.endswith(".cif") or inFileName.endswith(".mmcif"):
        cifToPdb(inFileName, outPDBFile)
        return outPDBFile
    else:
        print("ERROR (toPdb), Unknown file type for file = %s" % inFileName)


def toCIF(inFileName, outCIFFile):
    if inFileName.endswith(".cif") or inFileName.endswith(".mmcif"):
        return inFileName
    elif inFileName.endswith(".pdb") or inFileName.endswith(".ent"):
        pdbToCif(inFileName, outCIFFile)
        return outCIFFile
    else:
        print("ERROR (toCIF), Unknown file type for file = %s" % inFileName)


def getEnviron():
    environ = pwutils.Environ(os.environ)
    environ.update({'RCSBROOT': os.path.join(Plugin.getMaxitHome())
                    }, position=pwutils.Environ.REPLACE)
    environ.update({'PATH': os.path.join(Plugin.getMaxitHome(), 'bin')
                    }, position=pwutils.Environ.BEGIN)
    return environ


def _frombase(inFileName, outFileName, log, oParam=1):
    # check if maxit exists,
    # if it does not then complain
    # convert pdb to cif using maxit
    global maxitAvailable
    try:
        maxitAvailable
    except:
        if not os.path.exists(Plugin.getMaxitBin()):
            maxitAvailable = False
            # show error message
        else:
            maxitAvailable = True

    if maxitAvailable:
        args = ' -input "' + inFileName + '" -output "' + outFileName + \
               '" -o %d' % oParam
        log.info('Launching: ' + Plugin.getMaxitBin() + args)
        # run in the background
        env = getEnviron()
        pwutils.runJob(None, Plugin.getMaxitBin(), args, env=env)
    else:
        # this is not the ideal conversion but it is better
        # than nothing
        aSH = AtomicStructHandler()
        aSH.read(inFileName)
        aSH.write(outFileName)
        # show error message
        print(pwutils.redStr("Please, install maxit with the command 'scipion installb maxit'"))
        print(pwutils.redStr("and restart scipion. Packages bison and flex are needed."))
        print(pwutils.redStr("If maxit is installed check %s in scipion.conf" % MAXIT_HOME))


def fromPDBToCIF(inFileName, outFileName, log):
    _frombase(inFileName, outFileName, log, 1)


def fromCIFToPDB(inFileName, outFileName, log):
    _frombase(inFileName, outFileName, log, 2)


def fromCIFTommCIF(inFileName, outFileName, log):
    _frombase(inFileName, outFileName, log, 8)

def testLog(log, messages= None, sdterrLog = None):
    # Method to capture exceptions like error messages from Phenix when map and model are far apart
    if sdterrLog is None or messages is None:
        return

    # read all content of a file
    for line in sdterrLog(logFile=1):
        # check if string present in a file
        for message in messages:
            if message[0] in line:
                log.info(pwutils.redStr("Error:  Protocol aborted "))
                log.info(pwutils.redStr(f"Error: {message[1]}"))
                raise Exception(pwutils.redStr("Error: %s" % message[1]))
        if "Sorry" in line:
            log.info(pwutils.redStr("WARNING, %s" % line))

def retry(runEnvirom, program, args, cwd, listAtomStruct=[], log=None, clean_dir=None, messages=[], sdterrLog=None):
    messages.append(("Sorry: Input map is all zero after boxing", "Sorry: Input map is all zero after boxing"))

    try:
        runEnvirom(program, args, cwd=cwd)
    except:
        testLog(log, messages,sdterrLog)

        # first remove every files or directories in the extra folder,
        # except maps
        partiallyCleaningFolder(program, cwd)
        # something went wrong, may be bad atomStruct format
        log.info('retry with maxit conversion')

        for i, atomStructName in enumerate(listAtomStruct):
            if atomStructName.endswith(".pdb.cif"):
                aSH = AtomicStructHandler()
                aSH.read(atomStructName)
                aSH.write(atomStructName)
                try:
                    runEnvirom(program, args, cwd=cwd)
                except:
                    testLog(log, messages,sdterrLog)
            else:
                try:
                    if atomStructName.endswith(".pdb"):
                        newAtomStructName = os.path.abspath(
                            os.path.join(cwd, "retrypdb%d.cif" % i))
                        fromPDBToCIF(atomStructName, newAtomStructName, log)
                        _args = args.replace(atomStructName, newAtomStructName)
                        try:
                            runEnvirom(program, _args, cwd=cwd)
                        except:
                            testLog(log, messages,sdterrLog)
                            fromCIFTommCIF(newAtomStructName, newAtomStructName, log)
                            runEnvirom(program, _args, cwd=cwd)
                    elif atomStructName.endswith(".cif"):
                        newAtomStructName = os.path.abspath(
                            os.path.join(cwd, "retrycif%d.cif" % i))
                        fromCIFTommCIF(atomStructName, newAtomStructName, log)
                        _args = args.replace(atomStructName, newAtomStructName)
                        try:
                            runEnvirom(program, _args, cwd=cwd)
                        except:
                            testLog(log, messages,sdterrLog)
                            newAtomStructName = os.path.abspath(
                                os.path.join(cwd, "retrycif%d.pdb" % i))
                            fromCIFToPDB(atomStructName, newAtomStructName, log)
                            _args = args.replace(atomStructName, newAtomStructName)
                            runEnvirom(program, _args, cwd=cwd)
                            testLog(log, messages,sdterrLog)
                except:
                    testLog(log, messages,sdterrLog)
                    # first remove files or directories in the extra folder,
                    # except maps
                    partiallyCleaningFolder(program, cwd)
                    # biopython conversion
                    aSH = AtomicStructHandler()
                    if atomStructName.endswith(".pdb") or atomStructName.endswith(".ent"):
                        newAtomStructName = atomStructName.replace(".pdb", ".cif"). \
                            replace(".ent", ".cif")
                    else:
                        newAtomStructName = atomStructName
                    try:
                        aSH.read(atomStructName)
                        aSH.write(newAtomStructName)
                        _args = args.replace(atomStructName, newAtomStructName)
                        runEnvirom(program, _args, cwd=cwd)
                        testLog(log, messages,sdterrLog)
                    except:
                        print("CIF file standardisation failed.")
                        # atomStructName = newAtomStructName



# TODO this should no be here, ROB
def partiallyCleaningFolder(program, cwd):
    l1 = os.listdir(cwd)
    l2 = ['molprobity.out', 'molprobity_probe.txt', 'molprobity_coot.py']
    l3 = ['validation_cryoem.pkl']
    l4 = ['placed_model.pdb', 'placed_model.cif']

    for item in l1:
        if (program.endswith('real_space_refine.py') and
            (item.endswith('geo') or
             item.endswith('real_space_refined.log') or
             item.endswith('real_space_refine.pdb') or
             item.endswith('real_space_refined.cif'))) or \
                (program.endswith('molprobity.py') and item in l2) or \
                (program.endswith('validation_cryoem.py') and item in l3) or \
                (program.endswith('emringer.py') and
                 (item.endswith("emringer.csv") or
                  item.endswith("emringer.pkl") or
                  item.endswith("emringer_plots") or
                  item.startswith('emringer_transfer'))) or \
                (program.endswith('dock_in_map.py') and item in l4):
            path1 = os.path.join(cwd, item)
            if os.path.exists(path1):
                if os.path.isfile(path1) or os.path.islink(path1):
                    os.remove(path1)
                elif os.path.isdir(path1):
                    shutil.rmtree(path1)

def addScipionAttribute(cifDic, attributeScoresDic, attrName, recipient='residues'):
    '''Add a scipion attribute in a section of the CIF dictionary
    "cifDic": must be a cif dictionary parsed by Biopython
    "attributeScoresDic" a dictionary of the form {spec: value}"'''
    if not NAME in cifDic:
        cifDic[NAME], cifDic[RECIP] = [], []
        cifDic[SPEC], cifDic[VALUE] = [], []
    for spec in attributeScoresDic:
        value = attributeScoresDic[spec]
        cifDic[NAME].append(attrName)
        cifDic[RECIP].append(recipient)
        cifDic[SPEC].append(spec)
        cifDic[VALUE].append(str(value))
    return cifDic


# ***************************************************************************
# * Authors:    Daniel Marchán (da.marchan@cnb.csic.es)
# *
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# ***************************************************************************/
from pyworkflow.object import Pointer
from pyworkflow.tests import BaseTest, setupTestProject, DataSet
from pwem.protocols.protocol_import import ProtImportMicrographs
from pwem.protocols.protocol_boxsize_checkpoint import ProtBoxSizeCheckpoint


class TestBoxSizeCheckpoint(BaseTest):
    """ Test box size checkpoint protocol"""

    @classmethod
    def setUpClass(cls):
        setupTestProject(cls)


    def testBoxSizeCheckpoint(self):
        # Calculate a manual operation
        protNoInput = self._runBoxSizeCheckpointNoInputs(label='No inputs BoxSize checkpoint')
        self.assertEqual(protNoInput.boxSize, 300, "Computation without input does not match.")

        # Feed the protocol with boxSize1
        # Calculate an average and surpassing thresholds
        protWithInput = self.newProtocol(ProtBoxSizeCheckpoint,
                                        boxSize2= 400,
                                         boolBoxSize1=False,
                                        boolBoxSizeAvg= True)

        pntToOutput = Pointer(protNoInput, extended="boxSize")
        protWithInput.boxSize1.setPointer(pntToOutput)
        protWithInput.setObjLabel("Surpass threshold and average")
        self.launchProtocol(protWithInput)
        self.assertEqual(protWithInput.boxSize.get(), 350, "Computation without input does not match.")

        # Feed the protocol with boxSize2 input as well
        # Calculate with timer operation
        protWithBothInput = self.newProtocol(ProtBoxSizeCheckpoint,
                                             boolTimer=True,
                                             timeout="2m"
                                             )
        pntToOutput = Pointer(protNoInput, extended="boxSize")
        protWithBothInput.boxSize1.setPointer(pntToOutput)
        pntToOutput = Pointer(protWithInput, extended="boxSize")
        protWithBothInput.boxSize2.setPointer(pntToOutput)
        protWithBothInput.setObjLabel("2 inputs within threshoold and timer")
        self.launchProtocol(protWithBothInput)
        self.assertEqual(protWithBothInput.boxSize.get(), 300 ,"Computation without input does not match.")


    def _runBoxSizeCheckpointNoInputs(cls, label):
        protBoxSizeCheckpoint = cls.newProtocol(ProtBoxSizeCheckpoint,
                                            boxSize1= 300,
                                            boxSize2= 310)

        protBoxSizeCheckpoint.setObjLabel("Within threshold and stay with reference")
        cls.launchProtocol(protBoxSizeCheckpoint)

        return protBoxSizeCheckpoint
